#ifndef _ARRAY_HPP_
#define _ARRAY_HPP_

//#include <iostream.h>
#include <string.h>
#include <stdarg.h>

template<class type> type jMIN(type a,type b) {return (a<b)? a: b; };
template<class type> type jMAX(type a,type b) {return (a>b)? a: b; };
template<class type> type jSIGN	(const type& a) {return (a<0) ?   -1 :    1; };
template<class type> type jSIGN0	(const type& a) {
	if 			(a>0) return 1;
	else if	(a<0) return -1;
	else return 0;
};
template<class type> int  jROUND	(const type& a) {return (a>0) ? a+.5 : a-.5; };
template<class type> type jABS   (const type& a) {return (a>0) ?    a :   -a; };


bool array_output=true;

template <class type> class array {
	public:
		array<type>(void);
		virtual ~array<type>(void);					//virtual destructor (for matrix.h too)
		array<type>(int n,int *s,type f);		//ndims,size
		array<type>(const array<type>& a);	//cpy constructor
		array<type>(char *fn);							//read from file
		void store(char *fn);								//write to file
		void storeTxt(char *fn);								//write to file
		type token(void);										//periodic tokenizer (loops forever, if you let him)
		type token(int d);									//tokenizer -1 | 0 | 1 (actually takes any offset! non periodic!)
		type cycle(int d);									//sad, that we had to indroduce this function
		void record(type a);								//injects type at curr++ (abuses size[0] for current)
		array<int> sort(void);							//sorts array	and returns indices such that ???
		//type min(int &at);									//returns global min, at becomes position
		//type max(int &at);									//returns global max, at becomes position
		void lshift(int d);                 //d-th element is then at position 0, orientation remains.
		void unique(void);                  //matlab standard, see source code for details, implemented for efficiency
		array<int> find(void);							//matlab standard
		array<int> find(int n);							//for efficiency
		bool hasAt(type elm,int &pos);			//return true and data[pos]==elm, otherwise false
		void destruct(void) {for (int c1=0;c1<numel;c1++) delete data[c1];} //unsafe destructor
		array<type> queue(array<int> ind);
		array<type> queue1(array<int> ind);
		array<type> queue2(array<int> ind);
		void stretch(int n,type f);					//makes size[0]==n, filling new with f
		const array<type>& operator= (const array<type>& a);    //equal
		type& operator()(int c0,...);				//more flexible but probably slower
		type& operator[](const int& c0) {return data[c0];}; //good alternative		
		type* operator^(const int& c0) {return &data[c0];};				
		const array<type>& operator*=(type a);	//array*a
		const array<type>& operator+=(type a);	//array+a
		void apply(void (*func)(type));
		void apply(type (*func)(type));
		void applyAllPairs(void (*func)(type,type));
		bool all(void);
		type sum(void);
		bool operator==(const array<type>& a);
		array<type>& operator<=(type a);					//careful!
		array<type>& operator<=(array<type>& a);	//careful!
		array<type>& operator^=(const array<type>& a);
		array<type>& operator|=(array<type>& a);
		int ndims,numel;
		type *data;
		int *size; //size[0..ndims-1] is actual size | size[ndims] indexes current element for tokenizer
	private:
		void qsort(int a,int b,int *ind);
};
//constructors-------------------------------------------------------------
template <class type> array<type>::array(void) {data=NULL; size=NULL;};
template <class type> array<type>::~array(void) {
	if (data) {delete[] data; delete[] size;}
};
template <class type> array<type>::array(int n,int *s,type f) {
	int c1;
	ndims=n;
	size=new int[ndims+1];
	memcpy(size,s,4*ndims);	size[ndims]=0; //tokenizer prep
	numel=1; for (c1=0;c1<ndims;c1++) numel*=size[c1];	
	data=new type[numel];	
	for (c1=0;c1<numel;c1++)
	  data[c1]=f;
};
template <class type> array<type>::array(const array<type>& a) {
  int c1;
	ndims=a.ndims;
	numel=a.numel;
	size=new int[ndims+1];
	data=new type[numel];
	memcpy(size,a.size,4*(ndims+1));
	for (c1=0;c1<numel;c1++)
	  data[c1]=a.data[c1];
};
//file i/o------------------------------------------------------------------
template <class type> array<type>::array(char *fn) {//constructor!!!
	int c1;
	FILE *fid; //format: ndims,size[0:ndims-1],data
	if (fid=fopen(fn,"rb")) {
		fread(&ndims,4,1,fid);
		size=new int[ndims+1];
		fread(size,4,ndims,fid); size[ndims]=0; //tokenizer prep
		numel=1; for (c1=0;c1<ndims;c1++) numel*=size[c1];		
		//printf("numel %i  ", numel);
		//printf("ndims %i  ", ndims);
		//printf("s[0] %i  ", size[0]);
		//printf("s[1] %i  ", size[1]);
		//printf("s[2] %i  ", size[2]);
		//printf("s[3] %i  ", size[3]);
		//getchar();		
		data=new type[numel];
		fread(data,sizeof(type),numel,fid);
		fclose(fid);
		//if (array_output) cout << fn << " ->\t" << *this;
	} else {		
		data=NULL;
		//if (array_output) cout << "??? " << fn << " input suppressed!" << endl;
	}
};
template <class type> void array<type>::store(char *fn) {
	FILE *fid; //format: ndims,size[0:ndims-1],data
	if (fid=fopen(fn,"wb")) {
		fwrite(&ndims,4,1,fid);
		fwrite(size,4,ndims,fid);
		fwrite(data,sizeof(type),numel,fid);
		fclose(fid);
		//cout << fn << " <-\t" << *this;
	} //else cout << "??? " << fn << " output suppressed!" << endl;
}
//template <class type> void array<type>::storeTxt(char *fn) {
//	int c1,c2,c3;
//	ofstream fid(fn);
//  if (fid.is_open()) {	  
//	  fid << "% n-dimensional array file. n==2 corresponds to matrix or vector." << endl;
//	  fid << "% n | s1 s2 ... sn | data (filling up lower dimensions first = MATLAB standard)" << endl << endl;
//	  fid << ndims << endl;
//	  for (c1=0;c1<ndims;c1++) fid << size[c1] << ' ';
//	  fid << endl;
//	  for (c1=0;c1<numel;c1++) fid << data[c1] << ' ';
//	  fid << endl;
//  }
//  fid.close();
//  cout << fn << " <-T\t" << *this;
//}
template <class type> type& array<type>::operator()(int i0,...) {
	int c1,acc=1;
	va_list ap;
	va_start(ap,i0);
	for (c1=0;c1<ndims-1;c1++) {
		acc*=size[c1];
		i0+=va_arg(ap,int)*acc;
	}
	va_end(ap);	
	return data[i0];
};
//soft io funcs-------------------------------------------------------------
template <class type> type array<type>::token(void) {
	if (size[ndims]<numel) return data[size[ndims]++];
		else {size[ndims]=1; return data[0];}
};
template <class type> type array<type>::token(int d) {//change!!!
				if (size[ndims]>=numel) size[ndims]=numel-1;
	else 	if (size[ndims] <0)     size[ndims]=0;
	type ret=data[size[ndims]]; 	size[ndims]+=d;
	return ret;
};
template <class type> type array<type>::cycle(int d) {	//fundamental change!!!!
	type ret=data[size[ndims]]; 	size[ndims]+=d;         //size[ndims] has curtoken address
				if (size[ndims]>=numel) size[ndims]=0;
	else 	if (size[ndims] <0)     size[ndims]=numel-1;
	return ret;
};
template <class type> void array<type>::record(type a) {
	if (size[ndims]<numel) data[size[ndims]++]=a;
		else {size[ndims]=1; data[0]=a; }
};
//member funcs-------------------------------------------------------------
template <class type> void array<type>::apply(void (*func)(type)) {	
	for (int c1=0;c1<numel;c1++) func(data[c1]);
};	
template <class type> void array<type>::apply(type (*func)(type)) {	
	for (int c1=0;c1<numel;c1++) data[c1]=func(data[c1]);
};	
template <class type> void array<type>::applyAllPairs(void (*func)(type,type)) {
	int c1,c2;
	for (c1=0;c1<numel-1;c1++) 
		for (c2=c1+1;c2<numel;c2++)
			func(data[c1],data[c2]);
}
//advanced member funcs-------------------------------------------------------------
template <class type> array<int> array<type>::sort(void) {	
	array<int> ind(1,&numel,-1);
	for (int c1=0;c1<numel;c1++) ind.data[c1]=c1;
	qsort(0,numel-1,ind.data);
	return ind;
};
template <class type> void array<type>::qsort(int a,int b,int *ind) {
	int c1=a,c2=b,tpos;
	type ref=data[a],tdat; //selects first element!!!
	do {
		while (data[c1]<ref) c1++;
		while (data[c2]>ref) c2--;
		if (c1<=c2) {
					tpos=ind[c1];
			 ind[c1]=ind[c2];
			 ind[c2]=tpos;
					tdat=data[c1];
			data[c1]=data[c2];
			data[c2]=tdat;
			c1++;	c2--;
		}
	} while (c1<=c2);
	if (c2>a) qsort(a,c2,ind);
	if (c1<b) qsort(c1,b,ind);
};
/*
template <class type> type array<type>::min(int &at) {
	int c1;
	type cm=data[0];
	at=0;	
	for (c1=1;c1<numel;c1++)
		if (data[c1]<cm) {at=c1; cm=data[c1]; }
	return cm;
}
template <class type> type array<type>::max(int &at) {
	int c1;
	type cm=data[0];
	at=0;	
	for (c1=1;c1<numel;c1++)
		if (data[c1]>cm) {at=c1; cm=data[c1]; }
	return cm;
}
*/
template <class type> void array<type>::lshift(int d) {
	type* tdata=new type[numel];
	memcpy( tdata       ,&data[d],sizeof(type)*(numel-d));
	memcpy(&tdata[numel-d],data,sizeof(type)*d);
	delete[] data;
	data=tdata;
}
template <class type> void array<type>::unique(void) {
	int c1,cnt=1;	
	sort();
	type t0=data[0];
	for (c1=1;c1<numel;c1++)
		if (data[c1]!=t0) data[cnt++]=t0=data[c1];
	numel=cnt;	
};
template <class type> array<int> array<type>::find(void) {
	int cnt=0;
	for (int c1=0;c1<numel;c1++) if (data[c1]) cnt++;
	return find(cnt);
};
template <class type> array<int> array<type>::find(int n) {
	array<int> ind(1,&n,0);
	int cnt=0;
	for (int c1=0;c1<numel;c1++) 
		if (data[c1]&&cnt<n) ind.data[cnt++]=c1;
	return ind;
};
template <class type> bool array<type>::hasAt(type elm,int &pos) {//could make pos counter :-)
	int c1;
	for (c1=0;c1<numel;c1++)
		if (data[c1]==elm) { pos=c1; return true; }
	return false;
};
//operators----------------------------------------------------------------
template <class type> bool array<type>::operator==(const array<type>& a) {
	if (numel!=a.numel) return false;
	for (int c1=0;c1<numel;c1++)
		if (data[c1]!=a.data[c1]) return false;
	return true;
};
template <class type> const array<type>& array<type>::operator+=(type a) {
	for (int c1=0;c1<numel;c1++) data[c1]+=a;
	return *this;
};
template <class type> const array<type>& array<type>::operator*=(type a) {
	for (int c1=0;c1<numel;c1++) data[c1]*=a;
	return *this;
};
template <class type> array<type> array<type>::queue(array<int> ind) {
	array<type> ret(ind.ndims,ind.size,0);
	for (int c1=0;c1<ind.numel;c1++)
		ret.data[c1]=data[ind.data[c1]];
	return ret;
};
template <class type> array<type> array<type>::queue1(array<int> ind) {
	int c1,c2;
	int *sz=new int[ndims],ofs=0,cnt=0;
	memcpy(sz,size,4*ndims);
	sz[0]=ind.numel;
	array<type> ret(ndims,sz,0);
	for (c1=0;c1<numel;c1+=size[0]) 
		for (c2=0;c2<ind.numel;c2++) 
			ret.data[cnt++]=data[c1+ind[c2]];		
	return ret;
};
template <class type> array<type> array<type>::queue2(array<int> ind) {
	int c1;
	int *sz=new int[ndims];
	memcpy(sz,size,4*ndims);
	sz[1]=ind.numel;
	array<type> ret(ndims,sz,0);
	for (c1=0;c1<ind.numel;c1++)
		memcpy(&ret.data[c1*size[0]],&data[ind[c1]*size[0]],sizeof(type)*size[0]);	
	return ret;
};
template <class type> array<type>& array<type>::operator<=(type a) {
	data[size[ndims]++]=a;
	return *this;
}
template <class type> array<type>& array<type>::operator<=(array<type>& a) { //unsafe 
	memcpy(&data[size[ndims]],a.data,sizeof(type)*a.numel);
	size[ndims]+=a.numel;
	return *this;
}

template <class type> array<type>& array<type>::operator^=(const array<type> &a) {
	int c1,col=1,of1=0,of2=0,ofs=0;
	type* tdata=new type[numel+a.numel];
	for (c1=1;c1<ndims;c1++) col*=size[c1];
	for (c1=0;c1<col;c1++) {
		memcpy(&tdata[ofs],&  data[of1],sizeof(type)**  size); ofs+=  *size;
		memcpy(&tdata[ofs],&a.data[of2],sizeof(type)**a.size); ofs+=*a.size;
		of1+=  *size;
		of2+=*a.size;
	}
	delete[] data;
	data=tdata;
	*size+=*a.size;
	numel=*size*col;
	return *this;
}	

template <class type> array<type>& array<type>::operator|=(array<type> &a) {
	type* tdata=new type[numel+a.numel];	
	memcpy( tdata       ,  data,sizeof(type)*  numel);
	memcpy(&tdata[numel],a.data,sizeof(type)*a.numel);
	delete[] data;
	data=tdata;
	size[ndims-1]+=a.size[ndims-1];
	numel+=a.numel;
	return *this;
}

template <class type> void array<type>::stretch(int n,type f) {
	int c1,tnumel=n,rest=1;
	for (c1=1;c1<ndims;c1++) rest*=size[c1];
	tnumel*=rest;
	type* tdata=new type[tnumel];
	int cnt=0,buf=MAX(n-size[0],0);
	rest*=size[0];
	for (c1=0;c1<numel;c1+=size[0]) {
		memcpy(&tdata[cnt],&data[c1],sizeof(type)*size[0]);
		cnt+=n;
	}
	delete[] data;	
	data=tdata;
	numel=tnumel;
	size[0]=n;
}
//operators----------------------------------------------------------------
template <class type> const array<type>& array<type>::operator= (const array<type>& a) {					
	if (data==a.data) return *this;
	if (data) {delete[] data;	delete[] size;}
	ndims=a.ndims; size=new int[ndims+1];
	numel=a.numel; data=new type[numel];
	memcpy(size,a.size,4*(ndims+1));
	memcpy(data,a.data,sizeof(type)*numel);
	return *this;
};
template <class type> bool array<type>::all(void) {		
	for (int c1=0;c1<numel;c1++)
		if (!data[c1]) return false;
	return true;
};
template <class type> type array<type>::sum(void) {		
	type sum=0;
	for (int c1=0;c1<numel;c1++) sum+=data[c1];
	return sum;
};
template <class type> ostream& operator<< (ostream& os, array<type>& a);
template <class type> ostream& operator<< (ostream& os, array<type>& a) {
	int c1,c2;
	if (!array_output) return os;
	os << " [";
	if (!a.data) os << "empty]" << endl;	
	else {		
		for (c1=0;c1<a.ndims-1;c1++) os << a.size[c1] << "x";
		os << a.size[a.ndims-1] << "]=" << a.numel << "\tdata:";
		if (sizeof(type)==1) {os << "CHR" << endl; return os;} else os << a.data << endl;
		if (a.numel) {
			if (a.ndims==1)
				if (a.numel<=100) {
					for (c1=0;c1<a.size[0];c1++) os << a.data[c1] << "\t";
					os << endl;
				}
			if (a.ndims==2)
				if (a.size[0]<=500&&a.numel<=500)
					for (c1=0;c1<a.size[0];c1++) {
						for (c2=0;c2<a.size[1];c2++) os << a.data[c1+c2*a.size[0]] << "\t";
						os << endl;
					}
		}
	}
	return os;
};

//apply<float,double>(myFloatArray,sqrt);
template <class type,class conv> void apply(array<type>& a,conv (*func)(conv)) {	
	for (int c1=0;c1<a.numel;c1++)
		a.data[c1]=(type)func((conv)a.data[c1]);
}
/*template <class type> istream& operator>> (istream& is, array<type>& a);
template <class type> istream& operator>> (istream& is, array<type>& a) {
	char temp[MAXLEN];	
	is.getline(temp,MAXLEN);
	//a.cpy(temp);
	return is; };
*/
template <class type> void arrayTensor(void (*func)(type,type),array<type> &a,array<type> &b) {
	int c1,c2;
	for (c1=0;c1<a.numel;c1++)
		for (c2=0;c2<b.numel;c2++)
			func(a.data[c1],b.data[c2]);
}

#endif
