// code by christelle and jph

void cDocks(void) {
  matrix<float> tmp(3,3,"eye");
  
  tmp.data[7]=0.5f;      
  cDrop=cDock*tmp;        // used for patching the ground and preventing stones to be ...
  
  tmp.data[7]=1.00f;      // turning point is one meter from the dock
  cTurn=cDock*tmp;
}

// clears all the map
void cReset(void) {
  int c1;
  for (c1=0;c1<cDENSITY;c1++)
    cDensity[c1]=0;
  SDL_FillRect(cDensityDraw,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
  SDL_FillRect(cProfileDraw,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
  SDL_FillRect(cProfileGone,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
  SDL_FillRect(cRouteDraw  ,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
  cPos=matrix<float>(3,3,"eye");
  cPos.data[0]= 0;  cPos.data[3]=-1;  cPos.data[6]=0.5f;
  cPos.data[1]= 1;  cPos.data[4]= 0;  cPos.data[7]=1.0f;
  cInv=cPos.solve(matrix<float>(3,3,"eye"));
  cCompass.data[0]=cPos.data[3];
  cCompass.data[1]=cPos.data[4];
  matrix<float> tmp(3,3,"eye");  
  tmp.data[7]=-0.62f;
  cDock=cPos*tmp;
  cDocks();
  cStone.size[1]=0;
  cStone.numel  =0;
  jTic[jTOC_NEXT_PROFILE]=jCLOCK_FUNC+5000;       // profiling starts +5 sec from now
}

// updates position of robot
void cLocalize(void) {
  int c1,c2,c3;  
  int pix,piy,pin;
  matrix<float> ofs(3,3,"eye");
  matrix<float> frm;
  matrix<float> pnt;
  float ang,res,bst,cmp;
  int cnt,sum;  
  jTic[jTOC_TIME_LOCALIZE]=jCLOCK_FUNC;
  // BEGIN: find current orientation and position
  for (c1=0;c1<cRESOL;c1++) {
    matrix<float> mtc=cPos;         // initially best match is cPos, now look for better one
    res=cResol[c1];
    bst=5.0f;                       // threshhold for comparison
    for (c2=0;c2<cPARAM;c2+=4) {
      ang=cParam[c2  ]*res;
      ofs.data[0]= cos(ang);
      ofs.data[1]= sin(ang);
      ofs.data[3]=-ofs.data[1];
      ofs.data[4]= ofs.data[0];
      ofs.data[6]= cParam[c2+1]*res;
      ofs.data[7]= cParam[c2+2]*res;
      frm=cPos*ofs;
      pnt=frm*jLaser;
      sum=0;
      for (c3=0;c3<pnt.numel;c3+=3) {
        pix=(int)(jMAP_X+pnt.data[c3+1]*jGND2MAP);
        piy=(int)(jMAP_Y+pnt.data[c3  ]*jGND2MAP);
        pin=piy*jMAP_W+pix;
        if (0<pin&&pin<cDENSITY)
          sum+=cDensity[pin];
      }
      cmp=sum*0.005525f+cParam[c2+3];  // computes average plus penalty
      if (bst<cmp) {
        bst=cmp;
        mtc=frm;
      }
    }
    cPos=mtc;
  }
  int clk=jCLOCK_FUNC;
  jTic[jTOC_TIME_LOCALIZE]=clk-jTic[jTOC_TIME_LOCALIZE];
  // END: find current orientation and position
  // BEGIN: velocity computation  
  float dt=(clk-jTic[jTOC_LAST_SLAM])*0.001f;         // milliseconds to seconds  
  if (dt<0.12f) dt=0.12f;
  if (1.00f<dt) dt=1.00f;  
  float dti=1.0f/dt;
  matrix<float> mot=cInv*cPos;
  matrix<float> tmp(3,1);
  tmp.data[0]=dti*asin(mot.data[1]);
  tmp.data[1]=dti*     mot.data[6] ;
  tmp.data[2]=dti*     mot.data[7] ;
  for (c1=0;c1<3;c1++) {
    jVel.data[c1]*=0.66f;
    jVel.data[c1]+=0.34f*tmp.data[c1];
  }
  sprintf(jChar,  "t%6.3f /%6.3f",dt,dti);
  jDrawString(screen,jChar,126,0);
  jVelTip.data[0]=-jVel.data[0]*jTIP_OFFSET;
  jVelTip.data[1]= jVel.data[2];
  jVelTip.data[2]=0.0f;
  jTic[jTOC_LAST_SLAM]=clk;
  // END: velocity computation
}

// corrects numerically the cPos matrix and inverts
void cJustify(void) {
  float ang,def,cod,sid;
  matrix<float> nrm0(2,1),nrm1(2,1);
  nrm0.data[0]=cPos.data[0]; nrm0.data[1]=cPos.data[1]; nrm0.normalize(2);
  nrm1.data[0]=cPos.data[3]; nrm1.data[1]=cPos.data[4]; nrm1.normalize(2);  
  ang=acos(nrm0|nrm1);
  def=ang-1.5707963f;
  def*=0.5f;
  cod=cos(def);
  sid=sin(def);
  cPos.data[0]= cod*nrm0.data[0]-sid*nrm0.data[1];
  cPos.data[1]=+sid*nrm0.data[0]+cod*nrm0.data[1];
  cPos.data[3]= cod*nrm1.data[0]+sid*nrm1.data[1];
  cPos.data[4]=-sid*nrm1.data[0]+cod*nrm1.data[1];
  cPos.data[8]=1;
  cInv=cPos.solve(matrix<float>(3,3,"eye"));
}

// updates map at positions where necessary
// avoids scattered laser points
void cMapping(void) {
  int c1,c2;
  Uint32* myUint32;
  #define mNum 15
  float mRad[mNum]={.41,.35,.30,.25,.22,.19,.17,.15,.13,.11,.09,.07,.05,.03,.02}; // meters
  int   mCol[mNum]={  3,  4,  5,  6,  8, 10, 12, 14, 16, 18, 23, 20, 25, 29, 31};  
  SDL_FillRect(cDensityStub,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000)); // clear stub
  // BEGIN: prepare part of map determined by perceived data    
  matrix<float> tmp(2,1);                 // temp matrix to measure difference
  matrix<float> loc=jLaser;
  for (c1=0;c1<loc.size[1];c1++) {
    tmp.data[0]=loc.data[c1*3+0];
    tmp.data[1]=loc.data[c1*3+1];
    if (cDENSITY_RANGEMAX<tmp.normalize(2)) {
      loc.data[c1*3+0]=tmp.data[0]*0.2f;
      loc.data[c1*3+1]=tmp.data[1]*0.2f;
    }
  }  
  jDraw(jIN_DENSITYSTUB,cPos*loc,SDL_MapRGB(&jSDL_PixelFormat,000,000,001),&filledPolygonRGBA);     
  // BEGIN: insert intermediate points
  matrix<float> pnt=cPos*jLaser;          // employ filledPolygon as boolean mask      
  matrix<float> pni=pnt;                  // employ filledPolygon as boolean mask     
  for (c1=0;c1<pni.size[1];c1++) {        //insert intermediate points
    pni.data[c1*3+0]+=pni.data[c1*3+3];   //insert intermediate points
    pni.data[c1*3+1]+=pni.data[c1*3+4];   //insert intermediate points
    pni.data[c1*3+0]*=0.5f;               //insert intermediate points
    pni.data[c1*3+1]*=0.5f;               //insert intermediate points
  }
  // END: insert intermediate points
  // BEGIN: criteria to add
  matrix<bool> dif(1,jLaserSize);         // true if laser point contributes to map
  matrix<float> nrm(1,jLaserSize);        // distance of laser   
  //float nrm;
  int pix,piy,pin;
  nrm.data[0]=cDENSITY_RANGEMAX+0.1f;
  for (c1=1;c1<jLaserSize;c1++) {                                 // rate laser points
    dif.data[c1]=true;                              
    tmp.data[0]=jLaser.data[(c1-1)*3+0]-jLaser.data[c1*3+0];      // laser could be scattered 
    tmp.data[1]=jLaser.data[(c1-1)*3+1]-jLaser.data[c1*3+1];      // laser could be scattered 
    dif.data[c1]&=tmp.norm(2)<cDENSITY_JOIN;                      // laser could be scattered 
    tmp.data[0]=jLaser.data[c1*3+0];
    tmp.data[1]=jLaser.data[c1*3+1];
    nrm.data[c1]=tmp.norm(2);                                     // store distance of laserpoint
    dif.data[c1]&=nrm.data[c1]<cDENSITY_RANGEMAX;
    pix=(int)(jMAP_X+pnt.data[c1*3+1]*jGND2MAP);
    piy=(int)(jMAP_Y+pnt.data[c1*3+0]*jGND2MAP);
    pin=piy*640+pix;
    if (0<pin&&pin<cDENSITY)
      dif.data[c1]&=cDensity[pin]<29;                             // laser is already close to structure!
    else
      dif.data[c1] =false;
  }
  // END: criteria to add mountain area
  for (c1=0;c1<mNum;c1++)                             // loop all radii
    for (c2=1;c2<jLaserSize-1;c2++)                   // loop all laser points 0]  1 2 3 ... 179  [180      
      if (dif.data[c2]&&dif.data[c2+1]&&(cDENSITY_RANGEMIN<nrm.data[c2]||c1<5)) {   // 100% fine laser points or blurring effect
        jDraw(jIN_DENSITYSTUB,&pnt.data[c2*3],mRad[c1]+.001,SDL_MapRGBA(&jSDL_PixelFormat,000,000,mCol[c1],255),&filledEllipseRGBA);
        jDraw(jIN_DENSITYSTUB,&pni.data[c2*3],mRad[c1]+.001,SDL_MapRGBA(&jSDL_PixelFormat,000,000,mCol[c1],255),&filledEllipseRGBA);
      }      
  for (c2=1;c2<jLaserSize-1;c2++)                   // loop all laser points 0]  1 2 3 ... 179  [180      
    jDraw(jIN_PROFILEGONE,&pnt.data[c2*3],.30f,0xff008000,filledEllipseRGBA);
  // END: prepare part of map determined by perceived data
  // BEGIN: merge that 'part of map' into permanent map
  SDL_LockSurface(cDensityStub);              // to manipulate pixel array  
  myUint32=(Uint32*)(cDensityStub->pixels);   // to manipulate pixel array  
  int met;
  for (c1=0;c1<cDENSITY;c1++) {
    met=myUint32[c1]&0xff;
    if (cDensity[c1]<met)             // take maximum, either perceived or 
      cDensity[c1]=met;
  }
  SDL_UnlockSurface(cDensityStub);
  // END: merge that 'part of map' into permanent map
  // BEGIN: translate map into picture of map
  SDL_LockSurface(cDensityDraw);
  myUint32=(Uint32*)(cDensityDraw->pixels);
  for (c1=0;c1<cDENSITY;c1++)
    if (cDensity[c1])
      myUint32[c1]=SDL_MapRGBA(&jSDL_PixelFormat,000,cDensity[c1]<<3,000,255);
  SDL_UnlockSurface(cDensityDraw);
  // END: translate map into picture of map
}


// maps camera image to ground
void cProfile(void) {
  int c1,c2;
  int pix,piy,pin;
  if (jTog[jTOG_DOCK_IN_SIGHT]&&                            // are there any stone
      jTic[jTOC_NEXT_PTU]    <jTic[jTOC_LAST_SENSORS]&&     // is the PTU ready
      jBLURRED_FUNC
      //jVelTip.norm(2)        <jBLURRED
     ) {                   // are we too fast?
    cDock=cPos*jDock;
    cDocks();
    jDraw(jIN_PROFILEGONE,&cDrop.data[6],0.5f,0xff800080,filledEllipseRGBA);
    jDraw(jIN_PROFILEGONE,&cTurn.data[6],0.5f,0xff800080,filledEllipseRGBA);    
  }
  // BEGIN: match jStone into cStone
  if (                     0 <jStone.size[1]&&              // are there any stone
      jTic[jTOC_NEXT_PTU]    <jTic[jTOC_LAST_SENSORS]&&     // is the PTU ready
      jBLURRED_FUNC
      //jVelTip.norm(2)        <jBLURRED
     ) {                   // are we too fast?
	  float bst,cmp;
	  int mni;
	  matrix<float> tmp(2,1);
	  matrix<float> pnt=cPos*jStone;	                        // perceived stones in world coordinates
	  for (c1=0;c1<pnt.size[1];c1++) {                        // go thru the perceived stones
    	bst=100.f;                                            // assume: no match
    	for (c2=0;c2<cStone.size[1];c2++) {                   // find best match to each perceived stone
    	  tmp.data[0]=pnt.data[c1*3+0]-cStone.data[c2*3+0];
    	  tmp.data[1]=pnt.data[c1*3+1]-cStone.data[c2*3+1];
    	  cmp=tmp.norm(2);
      	if (cmp<bst) {
      	  bst=cmp;
      	  mni=c2;
    	  }
    	}
    	if (cSTONE_MATCH<bst) {                               // if best match is too far away
      	mni=cStone.size[1]++;
      	cStone.numel+=3;                                    // holy shit!
      	cStone.data[mni*3+0]=pnt.data[c1*3+0];
        cStone.data[mni*3+1]=pnt.data[c1*3+1];
        cStone.data[mni*3+2]=1;
  	  } else {
        cStone.data[mni*3+0]=cStone.data[mni*3+0]*.9f+pnt.data[c1*3+0]*.1f;               // average
        cStone.data[mni*3+1]=cStone.data[mni*3+1]*.9f+pnt.data[c1*3+1]*.1f;               // average
        lineRGBA(cMap,(Sint16)(jMAP_X+   pnt.data[ c1*3+1]*jGND2MAP),
                      (Sint16)(jMAP_Y+   pnt.data[ c1*3+0]*jGND2MAP),
                      (Sint16)(jMAP_X+cStone.data[mni*3+1]*jGND2MAP),
                      (Sint16)(jMAP_Y+cStone.data[mni*3+0]*jGND2MAP),255,192,000,255);        
      }
	  }
  }
  // END: match jStone into cStone
  // BEGIN: maps camera image to ground
  if (camera[jPROFILE_CAM]!=NULL&&
      jTic[jTOC_NEXT_PROFILE]<jTic[jTOC_LAST_SENSORS]&&
      jTic[jTOC_NEXT_PTU    ]<jTic[jTOC_LAST_SENSORS]&&
      jVelTip.norm(2)        <jBLURRED
     ) {
    SDL_FillRect(jProfileMask,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
    jDraw(jIN_PROFILEMASK,cPos*jSight[jPROFILE_CAM],SDL_MapRGB(&jSDL_PixelFormat, 64, 64,255),&filledPolygonRGBA);
    matrix<float> wld(3,1);
    matrix<float> loc(3,1);
    wld.data[2]=1;
    SDL_LockSurface(jProfileMask);
    SDL_LockSurface(cProfileDraw);
    SDL_LockSurface(camera[jPROFILE_CAM]);
    Uint32* myMask=(Uint32*)(jProfileMask->pixels);
    Uint32* myDraw=(Uint32*)(cProfileDraw->pixels);
    Uint32* myCamP=(Uint32*)(camera[jPROFILE_CAM]->pixels);
    for (c1=0;c1<cDENSITY;c1++)
      if (myMask[c1]&&!myDraw[c1]) {
        wld.data[0]=((c1/640)-jMAP_Y)*jMAP2GND;
        wld.data[1]=((c1%640)-jMAP_X)*jMAP2GND;
        loc=cInv*wld;
        pin=jGnd2cam(jPROFILE_CAM,loc.data);
        if (0<pin)
          myDraw[c1]=myCamP[pin];
      }
    SDL_UnlockSurface(camera[jPROFILE_CAM]);
    SDL_UnlockSurface(cProfileDraw);
    SDL_UnlockSurface(jProfileMask);
    jTic[jTOC_NEXT_PROFILE]=jTic[jTOC_LAST_SENSORS]+3000;
  }
  // END: maps camera image to ground
}

// cRouting computes from robots position distances to all points in wDensity (explored terrain)
// ONE evaluation of cRouting can take up to 2 SECONDS in computation time (usually < 1 SECOND however)
// COST FOR THE EDGES ARE NOT SYMMETRIC ! ! ! ! !
void cRouting(int lim) {
  int c1,c2,c3,c4;
  int pix,piy,pin,pia,pii;
  // BEGIN: clear
  for (c1=0;c1<cDENSITY;c1++) {
    cRouteDist[c1]=cDENSITY;
    cRoutePred[c1]=0;
    cRouteMask[c1]=0;
  }  
  // END: clear
  matrix<float> ahd(3,1);
  ahd.data[1]=0.13f;
  ahd.data[2]=1;
  ahd=cPos*ahd;
  pix=(int)(jMAP_X+ahd.data[1]*jGND2MAP);    
  piy=(int)(jMAP_Y+ahd.data[0]*jGND2MAP);    
  // BEGIN: do fancy shortest-path algorithm
  if (0<pix&&pix<jMAP_W&&0<piy&&piy<jMAP_H) {
    pia=piy*jMAP_W+pix;                               // map the robot into the map  
    jTic[jTOC_TIME_ROUTING]=jCLOCK_FUNC;
    linked<int>  list0,list1;
    linked<int>* now=&list0;
    linked<int>* nxt=&list1;
    linked<int>* tmp;
    now->append(pia);
    cRouteDist[pia]=0;
    cRoutePred[pia]=pia;
    float one=1.0f,rt2=sqrt(2.0);
    int   ofs[8]={-641,-639, 641, 639,  1,-640, -1,640};
    float fac[8]={ rt2, rt2, rt2, rt2,one, one,one,one};
    float met;
    int cnt=0;
    while (now->next[0]&&cnt<lim) {                   // repeat the process until now is empty, or mhh...
      cnt++;
      while (now->token(pia)) {                       // in the while loop, pia contains pixel address to be processed
        for (c1=0;c1<8;c1++) {
          pii=pia+ofs[c1];
          if (0<pii&&pii<cDENSITY)
            if (met=cDensity[pii]*fac[c1])            // only visit explored terrain (metric =0 whereever unexplored)
              if (cRouteDist[pia]+met<cRouteDist[pii]&&cRouteDist[pia]+met<lim) {    // IF ROUTING DOES NOT WORK, REMOVE latter condition!!!
                cRouteDist[pii]=cRouteDist[pia]+met;
                cRoutePred[pii]=pia;
                if (cRouteMask[pii]<cnt) {
                  cRouteMask[pii]=cnt;
                  nxt->append(pii);
                }
              }
        }
      }
      while (now->token(pia))                         // remove
        now->remove();
      tmp=now;
      now=nxt;
      nxt=tmp;
    }
    // END: do fancy shortest-path algorithm
    jTic[jTOC_TIME_ROUTING]=jCLOCK_FUNC-jTic[jTOC_TIME_ROUTING];
    // BEGIN: for visualization only
    SDL_FillRect(cRouteDraw,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
    SDL_LockSurface(cRouteDraw);
    Uint32* draw=(Uint32*)(cRouteDraw->pixels);
    for (c1=0;c1<cDENSITY;c1++)
      if (cRouteDist[c1]<cDENSITY) {
        draw[c1] =0xff000000;
        draw[c1]|=(((int)cRouteDist[c1])&0xff)<<1;
        //draw[c1]|=(cRouteMask[c1]&0xff)<<8;
      }
    SDL_UnlockSurface(cRouteDraw); 
    // END: for visualization only
  }
}

// function cPlanning starts from goal: cTarget and hops from predecessor to next predecessor 
// these pixel addresses are stored in a linked list (because the length of the path in pixels is
void cPlanning(void) {
  int c1,tok,pin,cnt=0;
  int ptx=(int)(jMAP_X+cTarget.data[7]*jGND2MAP); 
  int pty=(int)(jMAP_Y+cTarget.data[6]*jGND2MAP);
  int pta=pty*640+ptx;                              // pixel address of target  
  cPlanDist=cRouteDist[pta];                        // costs of the path are known from the beginning!
  while (cPlan.token(tok))                          // delete old path
    cPlan.remove();  
  while (cRoutePred[pta]&&pta!=cRoutePred[pta]) {   // hop from predecessor to next predecessor
    cPlan.append(pta);                              // until predecessor is NULL
    pta=cRoutePred[pta];
    cnt++;
  }
  array<int> pln(1,&cnt,0);                         //     
  c1=0;
  while (cPlan.token(tok)) {                        // delete old path
    pln.data[c1++]=tok;
    cPlan.remove();
  }
  cCompass*=0;  
  tok=0;
  for (c1=1;c1<=cnt;c1++) {
    cPlan.append(pin=pln.data[cnt-c1]);
    if (c1<10) {
      cCompass.data[0]+=((pin/640)-jMAP_Y)*jMAP2GND;
      cCompass.data[1]+=((pin%640)-jMAP_X)*jMAP2GND;
      tok++;
    }
  }
  if (3<tok) {
    cCompass.data[0]/=(float)tok;
    cCompass.data[1]/=(float)tok;
    cCompass.data[0]-=((pln.data[cnt-1]/640)-jMAP_Y)*jMAP2GND;
    cCompass.data[1]-=((pln.data[cnt-1]%640)-jMAP_X)*jMAP2GND;
    cCompass.data[2] =0;
    cCompass.normalize(2);
    matrix<float> loc=cInv*cCompass;
    float ang=atan2(-loc.data[0],loc.data[1]);    // angle with respect to front, left is positive
    if (100*jDEG2RAD<ang              )
      ang= 100*jDEG2RAD;
    if (             ang<-100*jDEG2RAD)
      ang=-100*jDEG2RAD;
    loc.data[0]=-sin(ang);
    loc.data[1]= cos(ang);
    loc.data[2]= 0;
    cCompass=cPos*loc;
  } else {
    cCompass.data[0]=cPos.data[3];
    cCompass.data[1]=cPos.data[4];
  }
}

int cBestQuality_PROTECTED(void) {
  int c1;
  SDL_LockSurface(cProfileDraw);                    // to manipulate pixel array  
  SDL_LockSurface(cProfileGone);
  Uint32* draw=(Uint32*)(cProfileDraw->pixels);   // to manipulate pixel array  
  Uint32* gone=(Uint32*)(cProfileGone->pixels);   // to manipulate pixel array  
  int pin=0,pix,piy;
  float cmp,bst=0;
  for (c1=0;c1<cDENSITY;c1++)
    if (cDensity[c1]==1&&draw[c1]==0&&gone[c1]==0&&cRouteMask[c1]) {
      cmp=cRouteDist[c1]*jRandu.token();
      if (bst<cmp) {
        bst=cmp;
        pin=c1;
      }
    }
  SDL_UnlockSurface(cProfileGone);
  SDL_UnlockSurface(cProfileDraw);  
  if (pin) {
    pix=pin%640;
    piy=pin/640;
    cTarget.data[6]=(piy-320)*jMAP2GND;
    cTarget.data[7]=(pix-320)*jMAP2GND;
    cTarget.data[8]=1;
    jDraw(jIN_PROFILEGONE,&cTarget.data[6],.35f,0xff800000,filledEllipseRGBA);
  }
  return pin;
}


// cPathHazardous returns false if the path is ok to drive
bool cPathHazardous(void) {
  int tok;
  bool dps=false;
  while (cPlan.token(tok))
    dps|=9<cDensity[tok];
  return dps;
}


int cBestQuality(void) {
  int pin;
  while (pin=cBestQuality_PROTECTED()) {
    if (!cPathHazardous())
      break;
  }
  return pin;
}
