// code by jph

// BEGIN: macros
#define           jCLOCK_FUNC         SDL_GetTicks()
#define           jLASER_OFFSET       0.105f  // offset of laser from center of robot
#define           jLASER_CUTOFF       6.0f    // 
#define           jYELLOW             15000
#define           jRELLOW             40
#define           jSTONES             10      // max number of stones per picture
#define           jROBOT_RADIUS_CM    22
#define           jROBOT_RADIUS_M     0.22f
#define           jDOCK_DEGREE        8
#define           jHAZARD_R0          0.290f
#define           jHAZARD_R1          0.420f
#define           jSPEED_ANG          0.60f    // max angular speed
#define           jSPEED_HED          0.25f   // 0.25 originally
#define           jBLURRED            0.12f    // dont take pictures if velocity is beyond that
#define           jSTILLSTAND         0.10f
#define           jTIP_OFFSET         0.4f
#define           jPROFILE_CAM        1
#define           jDOCK_CAM           1
#define           jCAMERA             76800
#define           jPROFILE            409600
#define           jFRONT              243200
#define           jDEG2RAD            0.0174533
#define           jRAD2DEG            57.295779
#define           jBLURRED_FUNC       fabs(jVel.data[0])<jBLURRED
#define           jEPS                0.0001
// END: macros

// BEGIN: constants
SDL_PixelFormat   jSDL_PixelFormat;           // pixel format for all SDL_Surfaces that we create
array<float>      jPTU_pan;                   // range for camera pan
array<float>      jPTU_til;                   // range for camera tilt
int               jFOV=0;                     // later the number of points of the field-of-view polygon
int               jFoV[128];                  // space for the field-of-view polygon
matrix<float>     jCamL2gnd(10,6);            // for camera <-> ground conversion
matrix<float>     jCamR2gnd(10,6);            // for camera <-> ground conversion
matrix<float>     jGnd2camL(10,6);            // for camera <-> ground conversion
matrix<float>     jGnd2camR(10,6);            // for camera <-> ground conversion
int               jLaserSize=181;             // number of laser range points
array<float>      jRandu(1,&jLaserSize,0);    // uniform random numbers accessed by periodic tokenizer
array<float>      jRandn(1,&jLaserSize,0);    // normal random numbers accessed by periodic tokenizer
// END: constants

// BEGIN: globals
char              jChar[250];
SDL_Surface*      jFront;                     // current view with respect to local j2b2 frame
SDL_Surface*      jProfileMask;
int               jCameraMask[jCAMERA];
sutree<int>*      jCameraTree[jCAMERA];
// END: globals

// BEGIN: sensors
matrix<float>     jLaser(3,jLaserSize);       // 181 laser points in local coordinates
matrix<float>     jSight[2];                  // true field-of-view in local coordinates
matrix<float>     jStone(3,jSTONES<<1);       // use jStone.size[1] to get number of stones that are perceived right now
matrix<float>     jReachable(3,jSTONES<<1);
matrix<float>     jDock(3,3,"eye");           // jDock in local coordinates if perceived
matrix<float>     jHazard(3,3,"eye");
float             jHazard_R0;                 // value determined by jHazards
float             jHazard_R1;                 // value determined by jHazards
float             jHazard_F0=0.4;             // used for bar
float             jHazard_F1=2.0;             // used for bar
// END: sensors

// BEGIN: actuators and sensors
matrix<float>     jPTU_old(2,1);              // stores old PTU to introduce delay if changed
matrix<float>     jSetSpeed(2,1);             // speed set by the planner
matrix<float>     jGetSpeed(2,1);             // final speed
float             jBoost=1.0f;
float             jAccel=0.2f;
float             jYellow=1.0f;
float             jRellow=1.0f;
float             jAvoid=1.0f;
matrix<float>     jVel(3,1);                  // this is local information, ccode writes into this!
matrix<float>     jVelTip(3,1);
// END: actuators

// BEGIN: internal timers all in MILLISECONDS
int               jTic[16];
#define           jTOC_LAST_SENSORS         0     // master counter
#define           jTOC_LAST_SLAM            1     // localize for speed
#define           jTOC_NEXT_PTU             2     // next PTU move
#define           jTOC_NEXT_PROFILE         3     // next profile
#define           jTOC_NEXT_SCREENSHOT      4     // next shot?
#define           jTOC_TIME_ROUTING         5
#define           jTOC_TIME_LOCALIZE        6
#define           jTOC_NEXT_STRATEGY        7     // next shot?
#define           jTOC_NEXT_SUBSTANTIAL     8
#define           jTOC_STONES_SINCE_VENTURE 9

bool              jTog[16];
#define           jTOG_DOCK_IN_SIGHT        0
#define           jTOG_IN_HAZARD            1
#define           jTOG_MIND_THE_GAP         2
#define           jTOG_PROFILE_IN_FRONTVIEW 3     // toggle ground vision in front view
#define           jTOG_CAMERA_BEHAVIOR      4
#define           jTOG_OBSTACLE_AVOIDANCE   5
#define           jTOG_ABORT_PATH_ON_STONE  6
#define           jTOG_ABORT_PATH_ON_DOCK   7
#define           jTOG_TRANSMIT_TO_GROUND1  8
// END: internals

#include "jsensor.cpp"

void jInit(void) {
  int c1;
  SDL_Surface* mySDL_Surface;

  #include "jconst.c"
  
  mySDL_Surface=SDL_LoadBMP("res/jfront.bmp");    // 380x640 black pixels
    jFront=SDL_ConvertSurface(mySDL_Surface,&jSDL_PixelFormat,SDL_SWSURFACE);
  SDL_FreeSurface(mySDL_Surface);
  
  mySDL_Surface=SDL_LoadBMP("res/cmap.bmp");      // 640x640    
    jProfileMask=SDL_ConvertSurface(mySDL_Surface,&jSDL_PixelFormat,SDL_SWSURFACE);    
    SDL_FillRect(jProfileMask,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
  SDL_FreeSurface(mySDL_Surface);
  
  wCam2gnd[0]=matrix<float>(3,5); wCam2gnd[0].data[14]=1;   // matrices for conversion
  wCam2gnd[1]=matrix<float>(3,5); wCam2gnd[1].data[14]=1;
  wGnd2cam[0]=matrix<float>(2,5);
  wGnd2cam[1]=matrix<float>(2,5);
  
  jSight[0]=matrix<float>(3,jFOV,"ones");
  jSight[1]=matrix<float>(3,jFOV,"ones");
  
  for (c1=0;c1<16;c1++) {
    jTic[c1]=0;
    jTog[c1]=false;
  }
  jTog[jTOG_PROFILE_IN_FRONTVIEW]=true;
  jTog[jTOG_CAMERA_BEHAVIOR     ]=true;
  jTog[jTOG_OBSTACLE_AVOIDANCE  ]=true;
  jTog[jTOG_TRANSMIT_TO_GROUND1 ]=true;
}

// manual steering
// manual control of PTU
void jProcess(SDL_Event &mySDL_Event) {
  if (mySDL_Event.type==SDL_KEYDOWN) {
    switch (mySDL_Event.key.keysym.sym) {
      case SDLK_UP:     jSetSpeed.data[0]= jSPEED_HED; break;
      case SDLK_DOWN:   jSetSpeed.data[0]=-jSPEED_HED; break;
      case SDLK_LEFT:   jSetSpeed.data[1]= jSPEED_ANG; break;
      case SDLK_RIGHT:  jSetSpeed.data[1]=-jSPEED_ANG; break;
      case SDLK_q:      ctrl.ASRobo_Stop(); break;        
    	case SDLK_f:      jPTU_pan.token(-1); break;
    	case SDLK_h:      jPTU_pan.token( 1); break;
    	case SDLK_t:      jPTU_til.token(-1); break;
    	case SDLK_g:      jPTU_til.token( 1); break;
    	case SDLK_e:      jTog[jTOG_PROFILE_IN_FRONTVIEW]=!jTog[jTOG_PROFILE_IN_FRONTVIEW]; break;
  	}
  }
  if (mySDL_Event.type==SDL_KEYUP) {
    switch (mySDL_Event.key.keysym.sym) {
      case SDLK_UP:     jSetSpeed.data[0]=0.0f; break;
      case SDLK_DOWN:   jSetSpeed.data[0]=0.0f; break;
      case SDLK_LEFT:   jSetSpeed.data[1]=0.0f; break;      
      case SDLK_RIGHT:  jSetSpeed.data[1]=0.0f; break;
  	}
  }
  if (mySDL_Event.type==SDL_MOUSEBUTTONDOWN) {
    int pix=mySDL_Event.button.x;
    int piy=mySDL_Event.button.y;
    if (jBAR_BOOST_X<pix&&pix<jBAR_BOOST_X+jBAR_W&&jBAR_BOOST_Y<piy&&piy<jBAR_BOOST_Y+jBAR_H)
      jBoost=(pix-jBAR_BOOST_X)*2.0f/(float)jBAR_W;
    if (jBAR_ACCEL_X<pix&&pix<jBAR_ACCEL_X+jBAR_W&&jBAR_ACCEL_Y<piy&&piy<jBAR_ACCEL_Y+jBAR_H) {
      jAccel=(pix-jBAR_ACCEL_X)*0.2f/(float)jBAR_W;
      //ctrl.ASRobo_SetAcceleration(jAccel);
    }
    if (jBAR_YELLOW_X<pix&&pix<jBAR_YELLOW_X+jBAR_W&&jBAR_YELLOW_Y<piy&&piy<jBAR_YELLOW_Y+jBAR_H)
      jYellow=(pix-jBAR_YELLOW_X)*2.0f/(float)jBAR_W;
    if (jBAR_RELLOW_X<pix&&pix<jBAR_RELLOW_X+jBAR_W&&jBAR_RELLOW_Y<piy&&piy<jBAR_RELLOW_Y+jBAR_H)
      jRellow=(pix-jBAR_RELLOW_X)*2.0f/(float)jBAR_W;
    if (jBAR_AVOID_X<pix&&pix<jBAR_AVOID_X+jBAR_W&&jBAR_AVOID_Y<piy&&piy<jBAR_AVOID_Y+jBAR_H)
      jAvoid=(pix-jBAR_AVOID_X)*2.0f/(float)jBAR_W;      
    if (jBAR_HAZARD0_X<pix&&pix<jBAR_HAZARD0_X+jBAR_W&&jBAR_HAZARD0_Y<piy&&piy<jBAR_HAZARD0_Y+jBAR_H)
      jHazard_F0=(pix-jBAR_HAZARD0_X)*3.0f/(float)jBAR_W;
    if (jBAR_HAZARD1_X<pix&&pix<jBAR_HAZARD1_X+jBAR_W&&jBAR_HAZARD1_Y<piy&&piy<jBAR_HAZARD1_Y+jBAR_H)
      jHazard_F1=(pix-jBAR_HAZARD1_X)*3.0f/(float)jBAR_W;
  }
}

// setSpeed of robot and setPTU
void jControl(void) {
  int c1,cnt;
  // BEGIN: obstacle avoidance camera behaviour
  if (jTog[jTOG_CAMERA_BEHAVIOR]&&
      jTic[jTOC_NEXT_PTU]<jTic[jTOC_LAST_SENSORS]
     ) {
    //if (jTog[jTOG_IN_HAZARD]) {
    //  if (0<jHazard.data[3]) {    // computes direction of hazard
    //    ang/=(float)cnt;
    //    if (0<ang)
    //      jPTU_pan.token(1);
    //    else
    //      jPTU_pan.token(-1);
    //} else {
    //  //go to old ptu
    //  jPTU_pan.size[1]=1;
    //}
  }
  // END: obstacle avoidance camera behaviour
  // BEGIN: PTU control
  if (jPTU_old.data[0]!=jPTU_pan.token(0)||jPTU_old.data[1]!=jPTU_til.token(0)) {
    jPTU_old.data[0]=jPTU_pan.token(0);
    jPTU_old.data[1]=jPTU_til.token(0);
    jTic[jTOC_NEXT_PTU]=jTic[jTOC_LAST_SENSORS]+2000;
  }
  ctrl.PTU_TurnTo(jPTU_pan.token(0),jPTU_til.token(0));
  // END: PTU control
  // BEGIN: speed control and obstacle avoidance
  float spd0,spd1;
  
  spd0=jSetSpeed.data[0]*jBoost+jRandu.token()*0.002f;  // boost and add noise to prevent motors to shut down
  spd1=jSetSpeed.data[1]*jBoost+jRandn.token()*0.002f;
  
  if (+jSPEED_HED<spd0) spd0= jSPEED_HED;               // make speed in the range!!
  if (spd0<-jSPEED_HED) spd0=-jSPEED_HED;
  if (+jSPEED_ANG<spd1) spd1= jSPEED_ANG;
  if (spd1<-jSPEED_ANG) spd1=-jSPEED_ANG;  
  
  float spg0=spd0;                                      // originals
  float spg1=spd1;                                      // originals
  
  if (jTog[jTOG_OBSTACLE_AVOIDANCE]&&jTog[jTOG_IN_HAZARD]) {
    float sca=7*jAvoid;
    matrix<float> nrm(3,1);
    nrm.data[0]=jHazard.data[3];
    nrm.data[1]=jHazard.data[4];
    sca*=nrm|jVelTip;                                   // project normal of hazard to tip velocity
    sprintf(jChar,"H |%6.3f",sca);
    jDrawString(screen,jChar,0,jBAR_HAZARD0_Y);    
    float cor;
    if (0<sca&&0.01<spd0) {                             // if robot is about to drive into wall, and wants to accelerate straight into wall (more than just noise)
      cor=MIN(sca,1.0f);
      if (0<jHazard.data[3])                            // depending on direction of Hazard...
        spd1= spd0*cor*jSPEED_ANG/jSPEED_HED;           // convert forward accel into sideward accel
      else
        spd1=-spd0*cor*jSPEED_ANG/jSPEED_HED;
      spd0*=(1-cor);                                    // by same factor reduce forward accel
    }
    if (0<sca&&spg1*jHazard.data[3]<0&&0.06<fabs(spg1)) {     // (more than just noise)
      spd0=-fabs(spg1)*jSPEED_HED/jSPEED_ANG*0.75f;
      jDrawString(screen,")|",70,jBAR_HAZARD0_Y);
    }
  }
  ctrl.ASRobo_SetSpeed(spd0,spd1);
  jGetSpeed.data[0]=spd0;
  jGetSpeed.data[1]=spd1;
  // END: speed control and obstacle avoidance
}

// contributes to camera
// line of straight ahead, pan, tilt
void jCameraPlot(void) {
  int c1,c2;
  int pin;
  int pix,piy;
  int prx,pry;
  bool lin;
  float pos[2];
  pos[0]=0;
  char lr[3]="LR";
  for (c1=0;c1<2;c1++)
    if (camera[c1]) {
      lin=false;
      for (c2=30;c2<=100;c2+=10) {    // draw straight line of heading
        pos[1]=c2*.01f;
        pin=jGnd2cam(c1,pos);          
        if (0<pin) {
          pix=pin%320;
          piy=pin/320;
          if (lin) {
            lineRGBA(camera[c1],pix,piy,prx,pry,jGRAY,jGRAY,jGRAY,255);
            prx=pix;
            pry=piy;
          } else {
            prx=pix;
            pry=piy;            
          }
          lin=true;
        } else
          lin=false;
      }
      sprintf(jChar,"%c",lr[c1]);     // label camera images left/right
      jDrawString(camera[c1],jFond,jChar,0,0);
    }
}

// contributes to front view
// line of straight ahead, pan, tilt
void jFrontPlot(void) {
  int c1;
  int pix,piy,pin,pif;
  SDL_Rect mySDL_Rect;
  mySDL_Rect.x=jFRONT_X;
  mySDL_Rect.y=0;
  mySDL_Rect.w=1;
  mySDL_Rect.h=640;
  SDL_FillRect(jFront,&mySDL_Rect,SDL_MapRGBA(&jSDL_PixelFormat,jGRAY,jGRAY,jGRAY,255));
  if (jTog[jTOG_IN_HAZARD]) {  // marks the circle redish if in hazard and plots the degree
    filledEllipseRGBA(jFront,jFRONT_X,jFRONT_Y,(Sint16)(jHazard_R1*jGND2FRONT),(Sint16)(jHazard_R0*jGND2FRONT),128, 64, 64,255);  
    filledEllipseRGBA(jFront,jFRONT_X,320,jROBOT_RADIUS_CM,jROBOT_RADIUS_CM,255, 64, 64,255);
    
    //sprintf(jChar,"%3.0f",jHazard*57.295779f);
    //jDrawString(jFront,jChar,jFRONT_X-30,jFRONT_Y-5);
  } else {
    filledEllipseRGBA(jFront,jFRONT_X,jFRONT_Y,(Sint16)(jHazard_R1*jGND2FRONT),(Sint16)(jHazard_R0*jGND2FRONT),128,128,128,255);  
    filledEllipseRGBA(jFront,jFRONT_X,jFRONT_Y,jROBOT_RADIUS_CM,jROBOT_RADIUS_CM,jGRAY,jGRAY,jGRAY,255);      
  }
  ellipseRGBA(jFront,jFRONT_X,jFRONT_Y,100,100,jGRAY,jGRAY,jGRAY,255); // plot radii of distance
  ellipseRGBA(jFront,jFRONT_X,jFRONT_Y,200,200,jGRAY,jGRAY,jGRAY,255);
  ellipseRGBA(jFront,jFRONT_X,jFRONT_Y,300,300,jGRAY,jGRAY,jGRAY,255);
  ellipseRGBA(jFront,jFRONT_X,jFRONT_Y,400,400,jGRAY,jGRAY,jGRAY,255);  
  
  // BEGIN: visualize speed vectors
  lineRGBA(jFront,jFRONT_X,jFRONT_Y,(Sint16)(jFRONT_X+jGND2FRONT*jVel.data[2]),
                                    (Sint16)(jFRONT_Y+jGND2FRONT*jVel.data[1]),000,255,000,255);  
  lineRGBA(jFront,(Sint16)(jFRONT_X+jGND2FRONT*(jTIP_OFFSET)),jFRONT_Y,(Sint16)(jFRONT_X+jGND2FRONT*(jVelTip.data[1]+jTIP_OFFSET)),
                                                                       (Sint16)(jFRONT_Y+jGND2FRONT*(jVelTip.data[0])),  0,255, 0,255);
  matrix<float> tmp(2,1);
  tmp.data[0]= jGetSpeed.data[0];
  tmp.data[1]=-jGetSpeed.data[1]*jTIP_OFFSET;  
  lineRGBA(jFront,(Sint16)(jFRONT_X+jGND2FRONT*(jTIP_OFFSET)),jFRONT_Y,(Sint16)(jFRONT_X+jGND2FRONT*(tmp.data[1]+jTIP_OFFSET)),
                                                                       (Sint16)(jFRONT_Y+jGND2FRONT*(tmp.data[0])),128,000,255,255);
  
  
  // END: visualize speed vector
  
  if (camera[jPROFILE_CAM]!=NULL&&
      jTog[jTOG_PROFILE_IN_FRONTVIEW]&&
      fabs(jVel.data[0])<jBLURRED
      //jVelTip.norm(2)<jBLURRED
     ) {
    SDL_FillRect(jProfileMask,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
    jDrawLocal(jIN_PROFILEMASK,jSight[jPROFILE_CAM],SDL_MapRGB(&jSDL_PixelFormat, 64, 64,255),&filledPolygonRGBA);
    matrix<float> loc(3,1);
    SDL_LockSurface(jProfileMask);
    SDL_LockSurface(jFront);
    SDL_LockSurface(camera[jPROFILE_CAM]);
    Uint32* myMask=(Uint32*)(jProfileMask->pixels);
    Uint32* myFron=(Uint32*)(jFront->pixels);
    Uint32* myCamP=(Uint32*)(camera[jPROFILE_CAM]->pixels);
    for (c1=0;c1<409600;c1++)
      if (myMask[c1]) {
        pix=c1%640;
        piy=c1/640;
        loc.data[0]=(piy-jFRONT_Y)*0.01f;
        loc.data[1]=(pix-jFRONT_X)*0.01f;
        pin=jGnd2cam(jPROFILE_CAM,loc.data);
        pif=piy*380+pix;
        if (0<pin&&0<pif&&pif<jFRONT)
          myFron[pif]=myCamP[pin];
      }
    SDL_UnlockSurface(camera[jPROFILE_CAM]);
    SDL_UnlockSurface(jFront); 
    SDL_UnlockSurface(jProfileMask); 
  }
}

void jHUD(void) {
  int c1,c2;
  Uint32 col;
  SDL_Rect mySDL_src;
  SDL_Rect mySDL_dst;
  mySDL_src.x=0; mySDL_src.y=0; mySDL_src.h=jBAR_H;
  // overall time
  mySDL_src.w=119;
  SDL_FillRect(screen,&mySDL_src,SDL_MapRGBA(screen->format,64,64,64,255));
  sprintf(jChar,"T %3i:%2i:%3i",
     jTic[jTOC_LAST_SENSORS]/60000,
    (jTic[jTOC_LAST_SENSORS]/1000)%60,
     jTic[jTOC_LAST_SENSORS]%1000);
  jDrawString(screen,jFond,jChar,0,0);
  // setSpeed
  sprintf(jChar,"S ^%6.3f >%6.3f",jSetSpeed.data[0]*jBoost,jSetSpeed.data[1]*jBoost);
  jDrawString(screen,jChar,0,jBAR_BOOST_Y);
  mySDL_src.w=(int)(jBAR_W*jBoost/2.0f);
  mySDL_dst.x=jBAR_BOOST_X;
  mySDL_dst.y=jBAR_BOOST_Y;  
  SDL_BlitSurface(jBarG,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%3.1f",jBoost);
	jDrawString(screen,jChar,jBAR_BOOST_X+jBAR_W/2-7,jBAR_BOOST_Y);
  // Accel
  mySDL_src.w=(int)(jBAR_W*jAccel/0.2f);
  mySDL_dst.x=jBAR_ACCEL_X;
  mySDL_dst.y=jBAR_ACCEL_Y;  
  SDL_BlitSurface(jBarG,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%3.1f",jAccel);
	jDrawString(screen,jChar,jBAR_ACCEL_X+jBAR_W/2-7,jBAR_ACCEL_Y);
  // camera Pan Tilt  
  sprintf(jChar,"C ^%6.3f >%6.3f ",jPTU_til.token(0),jPTU_pan.token(0));
  jDrawString(screen,jChar,0,jBAR_YELLOW_Y);
  // yellow
  mySDL_src.w=(int)(jBAR_W*jYellow/2.0f);
  mySDL_dst.x=jBAR_YELLOW_X;
  mySDL_dst.y=jBAR_YELLOW_Y;  
  SDL_BlitSurface(jBarY,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%3.1f",jYellow);
	jDrawString(screen,jChar,jBAR_YELLOW_X+jBAR_W/2-7,jBAR_YELLOW_Y);
  // rellow
  mySDL_src.w=(int)(jBAR_W*jRellow/2.0f);
  mySDL_dst.x=jBAR_RELLOW_X;
  mySDL_dst.y=jBAR_RELLOW_Y;  
  SDL_BlitSurface(jBarR,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%3.1f",jRellow);
	jDrawString(screen,jChar,jBAR_RELLOW_X+jBAR_W/2-7,jBAR_RELLOW_Y);
  sprintf(jChar,"V ^%6.3f >%6.3f *%6.3f\tv ^%6.3f >%6.3f |%6.3f",
    jVel.data[2],jVel.data[1],jVel.data[0],
    jVelTip.data[1],jVelTip.data[0],jVelTip.norm(2));
  jDrawString(screen,jChar,0,33);
  sprintf(jChar,"B c%i o%i",jTog[jTOG_CAMERA_BEHAVIOR],jTog[jTOG_OBSTACLE_AVOIDANCE]);
  jDrawString(screen,jChar,0,jBAR_AVOID_Y);
  // avoidance
  mySDL_src.w=(int)(jBAR_W*jAvoid/2.0f);
  mySDL_dst.x=jBAR_AVOID_X;
  mySDL_dst.y=jBAR_AVOID_Y;
  SDL_BlitSurface(jBarG,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%3.1f",jAvoid);
	jDrawString(screen,jChar,jBAR_AVOID_X+jBAR_W/2-7,jBAR_AVOID_Y);
	// hazard factor 0
  mySDL_src.w=(int)(jBAR_W*jHazard_F0/3.0f);
  mySDL_dst.x=jBAR_HAZARD0_X;
  mySDL_dst.y=jBAR_HAZARD0_Y;
  SDL_BlitSurface(jBarO,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%3.1f",jHazard_F0);
	jDrawString(screen,jChar,jBAR_HAZARD0_X+jBAR_W/2-7,jBAR_HAZARD0_Y);
	// hazard factor 0
  mySDL_src.w=(int)(jBAR_W*jHazard_F1/3.0f);
  mySDL_dst.x=jBAR_HAZARD1_X;
  mySDL_dst.y=jBAR_HAZARD1_Y;
  SDL_BlitSurface(jBarO,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%3.1f",jHazard_F1);
	jDrawString(screen,jChar,jBAR_HAZARD1_X+jBAR_W/2-7,jBAR_HAZARD1_Y);
}
