// code by jph

matrix<float> wCam2gnd[2];
// matrix contain ground coordinate corresponding to pixel pin in cam image
matrix<float> jCam2gnd(int cam,int pin) {
  float z0=(pin%320)-160; z0/=100.f;
  float z1=(pin/320)-120; z1/=100.f;
  matrix<float> piz(5,1);
    piz.data[0]=z0*z1;
    piz.data[1]=z1*z1;
    piz.data[2]=z0;
    piz.data[3]=z1;
    piz.data[4]=1;
  return wCam2gnd[cam]*piz;
}

matrix<float> wGnd2cam[2];
// returns number of pixel of gnd-coordinate in cam image
// returns -1     if coordinate is not within the cam image
int jGnd2cam(int cam,float* pos) {
  float z0=pos[0];
  float z1=pos[1]-0.20f; // sub 20 cm from center of robot to tip of robot front
  matrix<float> piz(5,1);
    piz.data[0]=z0*z1;
    piz.data[1]=z1*z1;
    piz.data[2]=z0;
    piz.data[3]=z1;
    piz.data[4]=1;
  matrix<float> ret=wGnd2cam[cam]*piz;
  int pix=(int)(ret.data[0]*100+160);
  int piy=(int)(ret.data[1]*100+120);
  if ( 0<pix&&pix<320 && 0<piy&&piy<240 )
    return piy*320+pix;
  else
    return -1;
}

// translate laser.distanceArray to local coordinates 
// camera <-> ground conversion matrices (left and right)
// compute field of view of camera[0] and camera[1] cut off by jLaser
void jSensors(void) {
  int c1,c2=0;
  // update internals
  jTic[jTOC_LAST_SENSORS]=jCLOCK_FUNC;
  // translate laser.distanceArray to local coordinates 
  double dst;
  double ang_abs=M_PI/2.;
  double ang_inc=M_PI/180.;
  for (c1=0;c1<jLaserSize;c1++) {
    dst=laser.distanceArray[c1].distance;
    if (jLASER_CUTOFF<dst)
      dst=laser.distanceArray[c1].distance=jLASER_CUTOFF;
    jLaser.data[c1*3+0]=sin(ang_abs)*dst;
    jLaser.data[c1*3+1]=cos(ang_abs)*dst+jLASER_OFFSET; // mounted offset from center
    jLaser.data[c1*3+2]=1;
    ang_abs-=ang_inc;
  }
  // camera <-> ground conversion matrices (left and right)
  float fp=jPTU_pan.token(0);
  float ft=jPTU_til.token(0);
  matrix<float> fpt(6,1);
    fpt.data[0]=fp*fp;
    fpt.data[1]=ft*fp;
    fpt.data[2]=ft*ft;
    fpt.data[3]=fp;
    fpt.data[4]=ft;
    fpt.data[5]=1;
  matrix<float> tmp;  
  tmp=jCamL2gnd*fpt; // left camera to ground
  for (c1=0;c1<5;c1++) {
    wCam2gnd[0].data[c1*3+0]=tmp.data[c1+0];
    wCam2gnd[0].data[c1*3+1]=tmp.data[c1+5];
  }
  wCam2gnd[0].data[13]+=0.20f; // add 20 cm from center of robot to tip of robot front  
  tmp=jCamR2gnd*fpt; // right camera to ground
  for (c1=0;c1<5;c1++) {
    wCam2gnd[1].data[c1*3+0]=tmp.data[c1+0];
    wCam2gnd[1].data[c1*3+1]=tmp.data[c1+5];
  }
  wCam2gnd[1].data[13]+=0.20f; // add 20 cm from center of robot to tip of robot front  
  tmp=jGnd2camL*fpt; // ground to left camera
  for (c1=0;c1<5;c1++) {
    wGnd2cam[0].data[c1*2+0]=tmp.data[c1+0];
    wGnd2cam[0].data[c1*2+1]=tmp.data[c1+5];
  }
  tmp=jGnd2camR*fpt; // ground to right camera
  for (c1=0;c1<5;c1++) {
    wGnd2cam[1].data[c1*2+0]=tmp.data[c1+0];
    wGnd2cam[1].data[c1*2+1]=tmp.data[c1+5];
  }  
  // compute field of view of camera[0] and camera[1] cut off by jLaser
  matrix<float> pos;
  int deg;
  for (c1=0;c1<2;c1++) { // both cameras
    for (c2=0;c2<jFOV;c2++) {
      pos=jCam2gnd(c1,jFoV[c2]);
      jSight[c1].data[c2*3+0]=pos.data[0];
      jSight[c1].data[c2*3+1]=pos.data[1];
      pos.data[1]-=.1f; // move 10 cm closer
      pos.data[2]=0;
      deg=(int)(atan2(-pos.data[0],pos.data[1])*57.295779f+90);
      if (0<deg&&deg<180)
        if (laser.distanceArray[deg].distance<pos.norm(2)) {
          jSight[c1].data[c2*3+0]=jLaser.data[deg*3+0];
          jSight[c1].data[c2*3+1]=jLaser.data[deg*3+1];
        }
    }
  }
}



// detect stones in a camera image, camera=camera[0] or camera=camera[1]
// writes pixels numbers into pin, at most jSTONES many
// returns number of stones
// implementation uses the disjoint-set union datastructure
// overall the performance is linear in the number of pixels
int jStones(SDL_Surface* camera,int* pins) {
  if (!camera)
    return 0;
  int c1;
  int RED,GRE,BLU;
  int cnt=0;                                  // number of yellow pixels
  int thr=(int)(jYELLOW*jYellow);
  for (c1=0;c1<jCAMERA;c1++)
    jCameraMask[c1]=0;
  SDL_LockSurface(camera);                    // to read and feedback into pixel array
  Uint32* myPixels=(Uint32*)(camera->pixels);  
  for (c1=320;c1<76800;c1++) {                // loop over all pixels and detect yellow    
    RED=myPixels[c1]&0xff0000; RED>>=16;
    GRE=myPixels[c1]&0xff00;   GRE>>=8;
    BLU=myPixels[c1]&0xff;
    if ( (255-abs(RED-GRE))*((RED+GRE)/2-BLU) > thr) {
      cnt++;
      jCameraMask[c1]=cnt;                    // store address instead of color! to identifz stones            
      myPixels[c1] =     BLU;                 // enhance pixel in image
      myPixels[c1]|=jMIN(GRE+50,255)<< 8;     //  ...
      myPixels[c1]|=jMIN(RED+50,255)<<16;     //  ...
      myPixels[c1]|=0xff000000;               //  ...
      jCameraTree[cnt]=new sutree<int>(c1);   // MAKESET, node has pixel address as content      
      if (jCameraMask[c1-1])                  // check pixel to the left
        jCameraTree[cnt]->join(jCameraTree[jCameraMask[c1-1]]);    // UNION
      if (jCameraMask[c1-320])                // check pixel one row above
        jCameraTree[cnt]->join(jCameraTree[jCameraMask[c1-320]]);  // UNION
    }
  }
  int pin,num=0;                              // number of stones
  for (c1=1;c1<=cnt;c1++) {                   // loop over all yellow pixels / nodes of union-data struct collecting unique representatives
    if (pin=jCameraTree[c1]->fset()->data) {  // pixel coordinate of representative node of stone    
      pins[num++]=pin;
      jCameraTree[c1]->fset()->data=0;
      if (num==jSTONES)
        break;
    }
  }
  SDL_UnlockSurface(camera);
  for (c1=1;c1<=cnt;c1++)
    delete jCameraTree[c1];                   // delete disjoint set union data struct
  return num;
}


// function jStones compiles the detected stones into
// the matrix jStone containing the local position of each stone
void jStones(void) {  
  int c1,c2;
  int num,deg,fin=0;
  int pins[jSTONES];    
  matrix<float> pos;
  for (c1=1;c1<2;c1++) {                      // which cameras to use: {0}, {1}, {0,1}
    num=jStones(camera[c1],pins);             // process camera image
    for (c2=0;c2<num;c2++) {
      pos=jCam2gnd(c1,pins[c2]);              // conversion formula for right camera
      pos.data[1]-=jLASER_OFFSET;             // compensate offset of laser
      pos.data[2] =0;                         // affine coordinate (remove for norm computation)
      deg=(int)(atan2(-pos.data[0],pos.data[1])*57.295779f+90); // number between 0 .. 180
      if (0<deg&&deg<180)
        if (pos.norm(2)<laser.distanceArray[deg].distance) {    // if stone is closer than laser
          jStone.data[fin*3+0]=pos.data[0];
          jStone.data[fin*3+1]=pos.data[1]+jLASER_OFFSET;
          jStone.data[fin*3+2]=1;
          fin++;                              // counts the stones
        }
    }
  }
  jStone.size[1]=fin;
  jStone.numel=3*fin;
}

void jReachables(void) {
  int c1,c2,cnt=0;  
  bool acc;
  float nrm;
  matrix<float> ray(2,1);
  matrix<float> pnt(2,1);
  for (c1=0;c1<jStone.size[1];c1++) {
    acc=false;
    ray.data[0]=-jStone.data[c1*3+1];   // rotate 90 degrees
    ray.data[1]= jStone.data[c1*3+0];
    nrm =ray.normalize(2);               // norm has the distance of stone
    nrm+=0.1f;
    for (c2=0;c2<jLaserSize;c2++) {
      pnt.data[0]=jLaser.data[c2*3+0];
      pnt.data[1]=jLaser.data[c2*3+1];
      acc|=fabs(ray|pnt)<0.25f&&pnt.norm(2)<nrm;
    }
    if (!acc) {
      jReachable.data[cnt*3+0]=jStone.data[c1*3+0];
      jReachable.data[cnt*3+1]=jStone.data[c1*3+1];
      jReachable.data[cnt*3+2]=1;
      cnt++;
    }
  }
  jReachable.size[1]=cnt;
  jReachable.numel=3*cnt;
}


// detect dock in a camera image
//    camera=camera[0] or camera[1]
//    if visualize is true, changes the image
// returns the pixel number where dock representative is
// if no dock in sight then returns 0
int jDocks(SDL_Surface* camera) {
  if (!camera) return 0;
  int c1;
  int RED,GRE,BLU;
  float prj0,prj1,prj2;
  int cnt=0; // number of yellow pixels
  Uint32 marker=SDL_MapRGB(&jSDL_PixelFormat,255,000,128);  
  SDL_LockSurface(camera);     // to manipulate pixel array  
  Uint32* myPixels=(Uint32*)(camera->pixels);
  float thr=jRELLOW*jRellow;
  for (c1=0;c1<jCAMERA;c1++)
    jCameraMask[c1]=0;
  for (c1=320;c1<76800;c1++) {        // loop over all pixels and detect yellow    
    RED=myPixels[c1]&0xff0000; RED>>=16;
    GRE=myPixels[c1]&0xff00;   GRE>>=8;
    BLU=myPixels[c1]&0xff;
    prj0=0.8473*RED+0.5311*BLU;
    prj1=(float)GRE;
    prj2=0.8473*BLU-0.5311*RED;
    if (thr<prj0-4*sqrt(prj1*prj1+prj2*prj2)) {
      cnt++;
      jCameraMask[c1]=cnt;                   // store address instead of color! to identifz stones            
      myPixels[c1]=marker;
      jCameraTree[cnt]=new sutree<int>(c1);  // MAKESET, node has pixel address as content      
      if (jCameraMask[c1-1])                 // check previous pixel
        jCameraTree[cnt]->join(jCameraTree[jCameraMask[c1-1]]);    // UNION
      if (jCameraMask[c1-320])               // check pixel one row above
        jCameraTree[cnt]->join(jCameraTree[jCameraMask[c1-320]]);  // UNION
    }
  }
  int pin=0; // number of stones
  if (cnt) {
    pin=jCameraTree[cnt]->fset()->data;    
    jDrawString(camera,"dock",280,0);
  }
  SDL_UnlockSurface(camera);
  for (c1=1;c1<=cnt;c1++)
    delete jCameraTree[c1]; // delete disjoint set union data struct
  return pin;
}


// this updates wDock if hole is detected
void jDocks(void) {
  int c1,c2,c3;
  int pin,deg;
  jTog[jTOG_DOCK_IN_SIGHT]=false;
  if (pin=jDocks(camera[jDOCK_CAM])) {                        // process image from right camera
    matrix<float> pos=jCam2gnd(jDOCK_CAM,pin);
    pos.data[1]-=jLASER_OFFSET;                               // move 10 cm closer
    pos.data[2] =0;                                           // position of perceived hole on ground with resp. to robot
    deg=(int)(atan2(-pos.data[0],pos.data[1])*57.295779f+90);
    if (20<deg&&deg<160)
      if (laser.distanceArray[deg].distance<pos.norm(2)) {    // if dock is perceived behind the wall
        jDock.data[6]=jLaser.data[deg*3+0];                   // dock is in local coordinates
        jDock.data[7]=jLaser.data[deg*3+1];
        matrix<float> tmp(2,1);
        tmp.data[0]=jLaser.data[(deg+jDOCK_DEGREE)*3+0]-jLaser.data[(deg-jDOCK_DEGREE)*3+0];
        tmp.data[1]=jLaser.data[(deg+jDOCK_DEGREE)*3+1]-jLaser.data[(deg-jDOCK_DEGREE)*3+1];
        tmp.normalize(2);
        jDock.data[0]= tmp.data[0];
        jDock.data[1]= tmp.data[1];
        jDock.data[3]=-tmp.data[1];
        jDock.data[4]= tmp.data[0];
        jTog[jTOG_DOCK_IN_SIGHT]=true;
      }
  }
}


// evaluates the laser scan to determine if robot is too close to obstacle
// function resets wHazard and jHazard
matrix<float> jHazardColl(2,jLaserSize);
void jHazards(void) {
  int c1,cnt=0;
  jHazard_R0=jHAZARD_R0;
  jHazard_R1=jHAZARD_R1;
  if (0<jVel.data[2]) {
    jHazard_R0+=jHazard_F0*jVel.data[2];  
    jHazard_R1+=jHazard_F1*jVel.data[2];
  }
  float iR0=1.0f/jHazard_R0;
  float iR1=1.0f/jHazard_R1;
  float a0,a1;
  jTog[jTOG_IN_HAZARD]=false;     // initially assume no hazard  
  for (c1=0;c1<jLaserSize;c1++) {
    a0=jLaser.data[c1*3+0]*iR0;
    a1=jLaser.data[c1*3+1]*iR1;
    if (a0*a0+a1*a1<1) {
      jHazardColl.data[cnt*2+0]=jLaser.data[c1*3+0];
      jHazardColl.data[cnt*2+1]=jLaser.data[c1*3+1];
      cnt++;
    }
  }
  if (5<cnt) {          // computes direction of hazard
    jTog[jTOG_IN_HAZARD]=true;
    matrix<float> tmp(2,1);    
    for (c1=0;c1<cnt;c1++) {
      tmp.data[0]+=jHazardColl.data[c1*2+0];
      tmp.data[1]+=jHazardColl.data[c1*2+1];
    }
    jHazard.data[6]=tmp.data[0]/(float)cnt;
    jHazard.data[7]=tmp.data[1]/(float)cnt;
    tmp.data[0]=jHazardColl.data[0]-jHazardColl.data[cnt*2-2];
    tmp.data[1]=jHazardColl.data[1]-jHazardColl.data[cnt*2-1];
    tmp.normalize(2);
    jHazard.data[0]= tmp.data[0];
    jHazard.data[1]= tmp.data[1];
    jHazard.data[3]=-tmp.data[1];
    jHazard.data[4]= tmp.data[0];
  }
}
