#ifndef _MATRIX_HPP_
#define _MATRIX_HPP_

//#include "array.hpp"
#define M_EPS 1E-20

template <class type> class matrix {
	public:
		matrix<type>(void);
		~matrix<type>(void);
		matrix<type>(const matrix<type> &arr);
		
  	const matrix<type>& operator= (const matrix<type>& a);    //equal

		//matrix<type>(char *fn) : array<type>(fn) {}; //read from file [from array.hpp]
		matrix<type>(int m,int n);              	//zeros(m x n)
		matrix<type>(int m,int n,char *c);				//various constructors		
		//matrix<type> diag(void);                	
		matrix<type> T(void);											//LATER TO BE replaced by bool ???
		matrix<type> solve(matrix<type> b);				//solve LGS
		matrix<type> inv2x2(void);								//invert 2x2 matrix
		type det2x2(void);
		type det3x3(void);
		matrix<type> glInverse(void);							//direct 4x4 openGL matrix inversion
		matrix<type> glDeriv(void);               // 4x4 with embedded SO(3)
		void glOrtho(void);							//direct 4x4 openGL matrix inversion
		void exp(void);
		type filter(matrix<type> *f);							//very spec but easy func (see explanation below)		
		type normalize(int n);										//normalize (vector) and return previous norm
		type norm(int n);													//n-norm of vector
		type norm(void);													//inf norm of matrix
		type mean(void);													//arithmetic mean
		matrix<type> operator+ (const matrix<type>& a); //matrix add
		matrix<type> operator- (const matrix<type>& a); //matrix-matrix
		matrix<type> operator* (const matrix<type>& a); //matrix mult		
		const matrix<type>& operator= (type a); 					//matrix=a
		matrix<type> operator+ (type a); 					//matrix+a
		matrix<type> operator- (type a); 					//matrix-a
		matrix<type> operator* (type a); 					//matrix*a
		matrix<type> operator% (const matrix<type>& a); //pointwise product!!!
		matrix<type> operator& (const matrix<type>& a); //cross product
		type         operator| (const matrix<type>& a); //scalar product
		matrix<type> operator> (type a); 					//matrix gt a
		const matrix<type>& operator+=(const matrix<type>& a);				//matrix+a		
		const matrix<type>& operator-=(const matrix<type>& a);				//matrix-a
		const matrix<type>& operator*=(const matrix<type>& a);				//matrix mult
		const matrix<type>& operator%=(const matrix<type>& a);				//pointwise prod
		const matrix<type>& operator&=(const matrix<type>& a);				//cross prod
		const matrix<type>& operator+=(type a);								//matrix+a
		const matrix<type>& operator-=(type a);								//matrix-a
		const matrix<type>& operator*=(type a);								//matrix*a
		const matrix<type>& operator/=(type a);								//a/matrix				
		//matrix<type>& operator<=(type a) {record(a); return *this;}
		const matrix<type>& operator<=(type a) {
  		data[size[ndims]++]=a;
  		return *this;
    } //unsafe but faster
		//void* operator[](int a) {return &data[a];};					//just to try!!		
		//matrix<type> operator[](const int& a);
		int ndims,numel;
		type *data;
		int *size; //size[0..ndims-1] is actual size | size[ndims] indexes current element for tokenizer
		
};
template <class type> matrix<type>::matrix(void) {
  data=NULL;
  size=NULL;
};
template <class type> matrix<type>::~matrix(void) {
	if (data) {
  	delete[] data;
  	delete[] size;
  }
};
template <class type> matrix<type>::matrix(const matrix<type>& a) {
  int c1;
	ndims=a.ndims;
	numel=a.numel;
	size=new int[ndims+1];
	data=new type[numel];
	memcpy(size,a.size,4*(ndims+1));
	for (c1=0;c1<numel;c1++)
	  data[c1]=a.data[c1];
};
template <class type> const matrix<type>& matrix<type>::operator= (const matrix<type>& a) {
	if (data==a.data) return *this;
	if (data) {
  	delete[] data;
  	delete[] size;
  }
	ndims=a.ndims; size=new int[ndims+1];
	numel=a.numel; data=new type[numel];
	memcpy(size,a.size,4*(ndims+1));
	memcpy(data,a.data,sizeof(type)*numel);
	return *this;
};




template <class type> matrix<type>::matrix(int m,int n) {	
	ndims=2; size=new int[3];
	size[0]=m; size[1]=n; size[2]=0; 
	numel=m*n;
	data=new type[numel];
	memset(data,0,sizeof(type)*numel);
}
template <class type> matrix<type>::matrix(int m,int n,char *c) {
	int c1;
	ndims=2; size=new int[3];
	size[0]=m; size[1]=n; size[2]=0; //tokenizer
	numel=m*n;
	data=new type[numel];
	if (!c[0]) 										memset(data,0,sizeof(type)*numel); //for (c1=0;c1<numel;c1++) data[c1]=0;
	if (!stricmp(c,"fill")) 			for (c1=0;c1<numel;c1++) data[c1]=c1;
	else if (!stricmp(c,"ones")) 	for (c1=0;c1<numel;c1++) data[c1]=1;
	else if (!stricmp(c,"eye")) {
		for (c1=0;c1<numel;c1++) data[c1]=0;
		for (c1=0;c1<numel;c1+=n+1) data[c1]=1;
	}
	else if (!stricmp(c,"nchoosek"))
		if (n==1) {
			data[0]=1;
			for (c1=1;c1<m;c1++) data[c1]=data[c1-1]*(m-c1)/c1;
		}
}
//:--------------------------------------------------------------
//template <class type> matrix<type> matrix<type>::diag(void) {	
//	matrix<float> x(numel,numel);
//	for (int c1=0;c1<numel*numel;c1+=numel+1) x.data[c1]=token();
//	return x;
//}
template <class type> matrix<type> matrix<type>::T(void) {
	int c1,c2,cnt=0;
	matrix<type> x(size[1],size[0]);
	for (c1=0;c1<size[1];c1++)
		for (c2=0;c2<numel;c2+=size[1])
		  x.data[c1+c2]=data[cnt++];
	return x;
}	
template <class type> matrix<type> matrix<type>::solve(matrix<type> b) {
	int c1,c2,c3,c1s=0,k;
	type tfac,tmax,tabs;	
	matrix<float> x(b.size[0],b.size[1]);
	matrix<int> p(size[0]+1,1,"fill");
	type *datacpy=new type[numel];
	memcpy(datacpy,data,sizeof(type)*numel);
	for (c1=0;c1<size[0];c1++) {
		k=c1; tmax=0;
		for (c2=c1;c2<size[0];c2++)
			if (tmax<(tabs=fabs(datacpy[p.data[c2]+c1s]))) {tmax=tabs; k=c2;}
		p.data[size[0]]=p.data[k];
					p.data[k]=p.data[c1];
				 p.data[c1]=p.data[size[0]];
		for (c2=c1+1;c2<size[0];c2++) {
			tabs=datacpy[p.data[c2]+c1s]/datacpy[p.data[c1]+c1s];
			for (c3=(c1+1)*size[0];c3<numel;c3+=size[0]) datacpy[p.data[c2]+c3]-=datacpy[p.data[c1]+c3]*tabs;
			for (c3=0;c3<b.numel;c3+=size[0]) b.data[p.data[c2]+c3]-=b.data[p.data[c1]+c3]*tabs;
		}
		c1s+=size[0];
	}
	for (c1=size[0]-1;c1>=0;c1--) {
		tfac=1/datacpy[p.data[c1]+c1*size[0]];
		for (c2=0;c2<b.size[1];c2++) {			
			tabs=b.data[p.data[c1]+c2*size[0]];			
			for (c3=c1+1;c3<size[0];c3++) 
				tabs-=datacpy[p.data[c1]+c3*size[0]]*x.data[c3+c2*size[0]];
			x.data[c1+c2*size[0]]=tabs*tfac;
		}
	}	
	delete[] datacpy;
	return x;
};
template <class type> matrix<type> matrix<type>::inv2x2(void) {
	matrix<float> x(2,2);
	float det=data[0]*data[3]-data[1]*data[2];	
	if (fabs(det)>M_EPS) {
		det=1/det;
		x.data[0]= data[3]*det;
		x.data[1]=-data[1]*det;
		x.data[2]=-data[2]*det;
		x.data[3]= data[0]*det;
	}
	return x;
};
template <class type> type matrix<type>::det2x2(void) {
	return data[0]*data[3]-data[1]*data[2];
};
template <class type> type matrix<type>::det3x3(void) {
	return -data[2]*data[4]*data[6]
	       +data[1]*data[5]*data[6]
	       +data[2]*data[3]*data[7]
	       -data[0]*data[5]*data[7]
	       -data[1]*data[3]*data[8]
	       +data[0]*data[4]*data[8];
};
template <class type> matrix<type> matrix<type>::glInverse(void) {	
	matrix<type> inv(4,4);	
	inv.data[0]=data[0]; inv.data[4]=data[1]; inv.data[ 8]=data[ 2];
	inv.data[1]=data[4]; inv.data[5]=data[5]; inv.data[ 9]=data[ 6];
	inv.data[2]=data[8]; inv.data[6]=data[9]; inv.data[10]=data[10];
	inv.data[12]-=data[12]*inv.data[0]+data[13]*inv.data[4]+data[14]*inv.data[ 8];
	inv.data[13]-=data[12]*inv.data[1]+data[13]*inv.data[5]+data[14]*inv.data[ 9];
	inv.data[14]-=data[12]*inv.data[2]+data[13]*inv.data[6]+data[14]*inv.data[10];
	inv.data[15]=1;
	return inv;
};
template <class type> matrix<type> matrix<type>::glDeriv(void) {
  matrix<type> mat(4,4);
  mat.data[1]= data[2];
  mat.data[2]=-data[1];
  mat.data[4]=-data[2];
  mat.data[6]= data[0];
  mat.data[8]= data[1];
  mat.data[9]=-data[0];
  return mat;
};
template <class type> void matrix<type>::glOrtho(void) {
	matrix<type> v0(3,1),v1(3,1),v2(3,1);
	memcpy(v0^0, data   ,12);
	memcpy(v1^0,&data[4],12);
	memcpy(v2^0,&data[8],12);
	v0.normalize(2);	
	v1-=v0*(v1|v0);
	v1.normalize(2);
	v2-=v0*(v2|v0);
	v2-=v1*(v2|v1);
	v2.normalize(2);
	memcpy( data   ,v0^0,12);
	memcpy(&data[4],v1^0,12);
	memcpy(&data[8],v2^0,12);
}
template <class type> void matrix<type>::exp(void) {
  int c1;  
  matrix<type> sum(size[0],size[1],"eye");
  matrix<type> nxt(size[0],size[1],"eye");
  for (c1=1;c1<40;c1++) {
    nxt=nxt* *this;
    nxt*=1/(type)c1;        
    sum+=nxt;
  }
  for (c1=0;c1<numel;c1++) 
    data[c1]=sum.data[c1];
};
//---------------------------
template <class type> type matrix<type>::filter(matrix<type> *f) {//this.records a arbitrary signal over the time (this is a "ring buffer") // f is a filter of the same numels as this
	int c1,cnt=0;	type sum=0;	
	for (c1=size[ndims];c1<numel;c1++) sum+=data[c1]*f->data[cnt++];
	for (c1=0;    c1<size[ndims];c1++) sum+=data[c1]*f->data[cnt++];
	return sum;
};
template <class type> type matrix<type>::norm(void) {
	int c1,c2;
	type mx,tmp;
	for (c1=0;c1<size[0];c1++) {
		tmp=0;
		for (c2=c1;c2<numel;c2+=size[1]) tmp+=jABS(data[c2]);
		mx=jMAX(mx,tmp);
	}
	return mx;	
};
template <class type> type matrix<type>::norm(int n) {
	int c1;
	type f=0;
	switch (n) {
	//case 0: //infinity norm
	//	for (c1=0;c1<numel;c1++) f=jMAX<type>(jABS(data[c1]),f);
	//	return f;
	//case 1:
	//	for (c1=0;c1<numel;c1++) f+=jABS(data[c1]);
	//	return f;
	case 2:
		for (c1=0;c1<numel;c1++) f+=data[c1]*data[c1];
		return sqrt(f);	
	}	
	return 0;
};
template <class type> type matrix<type>::normalize(int n) {
	int c1;
	type f=0,fi;
	switch (n) {
	//case 0: for (c1=0;c1<numel;c1++) f=jMAX<type>(jABS(data[c1]),f); break;
	case 1: for (c1=0;c1<numel;c1++) f+=jABS(data[c1]); break;		
	case 2: for (c1=0;c1<numel;c1++) f+=data[c1]*data[c1]; f=sqrt(f); break;
	}
	if (f>M_EPS) {
		fi=1/f;
		for (c1=0;c1<numel;c1++) data[c1]*=fi;
		return f;
	}	else return 0;
};
template <class type> type matrix<type>::mean(void) {
	type tmp=0;
	for (int c1=0;c1<numel;c1++) tmp+=data[c1];
	return tmp/numel;
};
//cunstructive operators-----------------------------------------------------------------------
template <class type> matrix<type> matrix<type>::operator+ (const matrix<type>& a) {
	matrix<type> b=*this;
	for (int c1=0;c1<MIN(numel,a.numel);c1++) b.data[c1]+=a.data[c1];
	return b; };
template <class type> matrix<type> matrix<type>::operator- (const matrix<type>& a) {
	matrix<type> b=*this;
	for (int c1=0;c1<MIN(numel,a.numel);c1++) b.data[c1]-=a.data[c1];
	return b; };
template <class type> matrix<type> matrix<type>::operator* (const matrix<type>& a) {
	int c1,c2,c3,s2=0,s3,as2=0;
	matrix<type> b(size[0],a.size[1]);
	for (c2=0;c2<a.size[1];c2++) {
		s3=0;
		for (c3=0;c3<size[1];c3++) {
			for (c1=0;c1<size[0];c1++)
				b.data[c1+s2]+=data[c1+s3]*a.data[as2+c3];					
			s3+=size[0];
		}
		s2+=size[0];
		as2+=a.size[0];
	}
	return b;
};

template <class type> const matrix<type>& matrix<type>::operator= (type a) {  
  for (int c1=0;c1<numel;c1++) data[c1]=a;
	return *this;
};
template <class type> matrix<type> matrix<type>::operator+ (type a) {
	matrix<type> b=*this;
	for (int c1=0;c1<numel;c1++) b.data[c1]+=a;
	return b; };
template <class type> matrix<type> matrix<type>::operator- (type a) {
	matrix<type> b=*this;
	for (int c1=0;c1<numel;c1++) b.data[c1]-=a;
	return b; };
template <class type> matrix<type> matrix<type>::operator* (type a) {
	matrix<type> b=*this;
	for (int c1=0;c1<numel;c1++) b.data[c1]*=a;
	return b; };
template <class type> matrix<type> matrix<type>::operator%(const matrix<type>& a) {
	matrix<type> b=*this;
	for (int c1=0;c1<MIN(numel,a.numel);c1++) b.data[c1]*=a.data[c1];
	return b; };	
template <class type> matrix<type> matrix<type>::operator& (const matrix<type>& a) {
	matrix<type> b(size[0],size[1]);
	b.data[0]=a.data[2]*data[1]-a.data[1]*data[2];
	b.data[1]=a.data[0]*data[2]-a.data[2]*data[0];
  b.data[2]=a.data[1]*data[0]-a.data[0]*data[1];
	return b; };
template <class type> type matrix<type>::operator|(const matrix<type>& a) {
	type sum=0;
	for (int c1=0;c1<MIN(numel,a.numel);c1++) sum+=data[c1]*a.data[c1];
	return sum; };
template <class type> matrix<type> matrix<type>::operator> (type a) {
	matrix<type> b(size[0],size[1]);
	for (int c1=0;c1<numel;c1++) b.data[c1]=(data[c1]>a);
	return b; };
//preserving operators-------------------------------------------------------------------------
template <class type> const matrix<type>& matrix<type>::operator+=(const matrix<type>& a) {
	for (int c1=0;c1<MIN(numel,a.numel);c1++) data[c1]+=a.data[c1];
	return *this;
};
template <class type> const matrix<type>& matrix<type>::operator-=(const matrix<type>& a) {
	for (int c1=0;c1<MIN(numel,a.numel);c1++) data[c1]-=a.data[c1];
	return *this;
};
template <class type> const matrix<type>& matrix<type>::operator*=(const matrix<type>& a) {
	int c1,c2,c3,s2=0,s3,as2=0; //NOTE: data!=a.data
	type *datacpy=new type[numel];
	memcpy(datacpy,data,sizeof(type)*numel);
	memset(data,0,sizeof(type)*numel);
	for (c2=0;c2<a.size[1];c2++) {
		s3=0;
		for (c3=0;c3<size[1];c3++) {
			for (c1=0;c1<size[0];c1++)
				data[c1+s2]+=datacpy[c1+s3]*a.data[as2+c3];			
			s3+=size[0];
		}
		s2+=size[0];
		as2+=a.size[0];
	}
	size[1]=a.size[1];
	numel=size[0]*size[1];
	delete[] datacpy;
	return *this;
};
template <class type> const matrix<type>& matrix<type>::operator%=(const matrix<type>& a) {
	for (int c1=0;c1<MIN(numel,a.numel);c1++) data[c1]*=a.data[c1];
	return *this;
};	
template <class type> const matrix<type>& matrix<type>::operator+=(type a) {
	for (int c1=0;c1<numel;c1++) data[c1]+=a;
	return *this;
};
template <class type> const matrix<type>& matrix<type>::operator-=(type a) {
	for (int c1=0;c1<numel;c1++) data[c1]-=a;
	return *this;
};
template <class type> const matrix<type>& matrix<type>::operator*=(type a) {
	for (int c1=0;c1<numel;c1++) data[c1]*=a;
	return *this;
};
template <class type> const matrix<type>& matrix<type>::operator/=(type a) {
	for (int c1=0;c1<numel;c1++) data[c1]=a/data[c1];
	return *this;
};
template <class type> const matrix<type>& matrix<type>::operator&=(const matrix<type>& a) {
	type t0,t1;
	t0=a.data[2]*data[1]-a.data[1]*data[2];
	t1=a.data[0]*data[2]-a.data[2]*data[0];
  data[2]=a.data[1]*data[0]-a.data[0]*data[1];	
  data[0]=t0;
  data[1]=t1;
	return *this;
};
//specific functions-----------------------------------------------------------------
//OPENGL standard
//-----------------------------------------------------------------------------------
matrix<float> glRotation(float *a) {//Rx.Ry.Rz
	matrix<float> rot(4,4);
	float sina=sin(a[0]),cosa=cos(a[0]),
				sinb=sin(a[1]),cosb=cos(a[1]),
				sinc=sin(a[2]),cosc=cos(a[2]);
	rot.data[0]= cosb*cosc;									rot.data[4]=-cosb*sinc;									rot.data[ 8]= sinb;
  rot.data[1]= cosc*sina*sinb+cosa*sinc;  rot.data[5]= cosa*cosc-sina*sinb*sinc;	rot.data[ 9]=-cosb*sina;
  rot.data[2]=-cosa*cosc*sinb+sina*sinc;	rot.data[6]= cosc*sina+cosa*sinb*sinc;	rot.data[10]= cosa*cosb;    
	rot.data[15]=1;
  return rot;
}
matrix<float> glRotation(float ax,float ay,float az) {
	float a[3]; a[0]=ax; a[1]=ay; a[2]=az;
	return glRotation(a);
}

matrix<float> glTranslate(float a,float b,float c) {//who did this
	matrix<float> tra(4,4,"eye");
	tra.data[12]=a; tra.data[13]=b; tra.data[14]=c;
	return tra;
}
matrix<float> glTranslate(float *a) {
	matrix<float> tra(4,4,"eye"); memcpy(&tra.data[12],a,12);
	return tra;
}
matrix<float> glTranslateInv(float *a) {
	matrix<float> tra(4,4,"eye");
	tra.data[12]=-a[0];
	tra.data[13]=-a[1];
	tra.data[14]=-a[2];
	return tra;
}
matrix<float> glScale(float a,float b,float c) {
	matrix<float> tra(4,4,"eye");
	tra.data[ 0]=a; tra.data[ 5]=b; tra.data[10]=c;
	return tra;
}
matrix<float> glScale(float a) {
	matrix<float> tra(4,4,"eye");
	tra.data[ 0]=a; tra.data[ 5]=a; tra.data[10]=a;
	return tra;
}
matrix<float> glEliminateY(matrix<float>& mat) {
	matrix<float> ref(4,4,"eye");
	matrix<float> nse(4,1);
	memcpy(nse.data,&mat.data[8],12);
	nse.data[1]=0;
	if (nse.normalize(2)>M_EPS) {
		memcpy(&ref.data[8],nse.data,12);
		ref.data[0] =ref.data[10];
		ref.data[2]-=ref.data[ 8];
	}
	memcpy(&ref.data[12],&mat.data[12],12);
	return ref;
}

matrix<float> glIntegral(float dt,float my,float la) {
	matrix<float> x(3,3);
	x.data[0]=1;
	x.data[3]=  dt;     	x.data[6]=  dt*dt*.5;
	x.data[4]=1-dt*my;    x.data[7]=  dt;
												x.data[8]=1-dt*la;
	return x;
}

template <class type> void glAverageNormals(array<float> *v,array<type> *t,bool annul) {
	int c1,c2,c3;
	int cnt=v->size[0];
	if (annul) for (c1=0;c1<v->numel;c1+=cnt) v->data[c1+cnt-6]=v->data[c1+cnt-5]=v->data[c1+cnt-4]=0;
	matrix<float> nrm(3,t->size[1]),v0(3,1),v1(3,1),v2(3,1);	
	for (c1=0;c1<t->numel;c1+=3) {
		memcpy(v0.data,&v->data[cnt*t->data[c1  ]+cnt-3],12);
		memcpy(v1.data,&v->data[cnt*t->data[c1+1]+cnt-3],12);
		memcpy(v2.data,&v->data[cnt*t->data[c1+2]+cnt-3],12);
		v1-=v0;
		v2-=v0;
		v1&=v2;
		v1.normalize(2);
		memcpy(&nrm.data[c1],v1.data,12);
	}
	for (c1=0;c1<t->numel;c1+=3) //for all triangles T
		for (c2=0;c2<3;c2++)       //for the 3 vertices of t
			for (c3=0;c3<3;c3++)     //for the xyz coordinates
				v->data[cnt*t->data[c1+c2]+cnt-6+c3]+=nrm.data[c1+c3];
	for (c1=0;c1<v->numel;c1+=cnt) {
		memcpy(v0.data,&v->data[c1+cnt-6],12);
		v0.normalize(2);
		memcpy(&v->data[c1+cnt-6],v0.data,12);
	}	
}

//matrix<float> glFull(float *v0,float *v1) {
//	int c1;
//	matrix<float> ret=glTranslate(v0);
//	matrix<float> vc0(3,1),vc1(3,1);
//	memcpy(vc0.data,v0,12);
//	memcpy(vc1.data,v1,12);
//	matrix<float> mzt=vc0-vc1;
//	float len=mzt.normalize(2);	
//	memcpy(ret^8,(mzt*len)^0,12);
//	matrix<float> eye(3,3,"eye");
//	matrix<float> res=eye-mzt*mzt.T();
//	for (c1=0;c1<3;c1++) {
//		memcpy(vc0.data,res^4*c1,12);
//		if (vc0.norm(2)>.1) {			
//			vc0-=mzt*(mzt|vc0);
//			vc0.normalize(2);			
//			memcpy(ret^0,vc0^0,12);
//			vc0&=mzt;
//			memcpy(ret^4,vc0^0,12);
//			break;
//		}
//	}	
//	return ret;
//}

//-------------------------------------------------------------------------------------
// TRIDIAGONAL SOLVER
//-------------------------------------------------------------------------------------
template <class type> class tridiag : public matrix<type> {
	public:
		tridiag<type>(void) {};
		matrix<type> solve(matrix<type> *b);			//solve LGS
	matrix<type> *l,*d,*r;											//left diag right (formats dont matter 0..need)
};

template <class type> matrix<type> tridiag<type>::solve(matrix<type> *b) {
	int c1;
	matrix<type> x(b->size[0],b->size[1]);
	matrix<type> g(b->size[0],b->size[1]);
	type beta=1/d->data[0];
	x.data[0]=b->data[0]*beta;
	for (c1=1;c1<b->numel;c1++) {
		g.data[c1]=r->data[c1-1]*beta;
		x.data[c1]=(b->data[c1]-l->data[c1-1]*x.data[c1-1])*(beta=1/(d->data[c1]-l->data[c1-1]*g.data[c1]));
	}
	for (c1=b->numel-1;c1>0;c1--)
		x.data[c1-1]=x.data[c1-1]-g.data[c1]*x.data[c1];
	return x;
};

//-------------------------------------------------------------------------------------
// SPARSE MATRIX
//-------------------------------------------------------------------------------------
template <int key,class type> class sparse {
	public:
		sparse(void);		
		sparse(char *fn);        //contruct from files i.e. 2 files so omit extention
		sparse(matrix<type> &a); //make full matrix sparse
		~sparse(void);
		matrix<type> multMatSparse(matrix<type> &a); // A * Sparse
		array<unsigned int> *ind;
		array<type> *val;
		int size[2];
};
template <int key,class type> sparse<key,type>::sparse(void) {
	ind=NULL; val=NULL;
};
template <int key,class type> sparse<key,type>::sparse(matrix<type> &a) {
	int c1,c2,ofs=0;
	array<int> ptr=a.find();
	int sz[2]; sz[0]=2; sz[1]=ptr.numel;	
	ind=new array<unsigned int>(2,sz,0);
	val=new array<type>(1,&ptr.numel,0);
	for (c1=0;c1<ptr.numel;c1++) {
		ind->data[c1*2+0]=ptr[c1]%a.size[0];
		ind->data[c1*2+1]=ptr[c1]/a.size[0];
		val->data[c1]		 =a[ptr[c1]];
	}
	*ind*=key;
	memcpy(size,a.size,8);
}
template <int key,class type> sparse<key,type>::~sparse(void) {
	if (ind) {delete ind; delete val;}	
}
template <int key,class type> matrix<type> sparse<key,type>::multMatSparse(matrix<type> &a) {
	int c1,c2,cnt=0,i,j;
	matrix<float> x(a.size[0],size[1]);
	for (c1=0;c1<val->numel;c1++) {
		i=ind->data[cnt++];
		j=ind->data[cnt++];
		for (c2=0;c2<a.size[0];c2++)
			x.data[c2+j]+=a.data[c2+i]*val->data[c1];	
	}
	return x;
}

#endif
