// code by paavo and all others

#define       pMASTER_NO_AUTONOMY               0
#define       pMASTER_TURN_LEFT_THEN_VENTURE    1
#define       pMASTER_VENTURE                   2
#define       pMASTER_DOCKING1_OLD_TURN_POINT   3
#define       pMASTER_DOCKING2_SAFEGUARD        4
#define       pMASTER_DOCKING3_KICKASS          5
#define       pMASTER_DEBUG1                    9
                                      
#define       pSTATE_IDLE                       0       // == NULL
#define       pSTATE_ALIGN_TO_COMPASS           1
#define       pSTATE_DRIVE_ALONG_PATH           2
#define       pSTATE_TRACK_STONE_GRAB           3
#define       pSTATE_VIVE_LA_RESISTANCE         4
#define       pSTATE_FUBAR                      5
#define       pSTATE_CLEAR_PROFILE              8
#define       pSTATE_SLAM_RESET                 9
              
#define       pSTATES                           20      // maximum numbers of states in sequence
                                     
#define       pGAININT                          1.0f    // previously we had an I-controller
#define       pGAINPOS                          1.0f    // now we have a PI-controller
float         pGainInt=0.80f;
float         pGainPos=0.70f;

#define       pTIMEOUT                          80000   // in milli seconds

// BEGIN: globals
int           pMaster;
int           pMasterOld=0;
int           pStates[pSTATES];                         // sequence of elementary states
int           pIndex=0;
// END: globals

#include "pmagic.c"

void pInit(void) {
  int c1;
  pMaster=pMASTER_NO_AUTONOMY;
  pStates[0]=pSTATE_IDLE;
}

void pProcess(SDL_Event &mySDL_Event) {
  int c1;
  if (mySDL_Event.type==SDL_KEYDOWN) {
    switch (mySDL_Event.key.keysym.sym) {
      case SDLK_0: {
        pMaster=pMASTER_NO_AUTONOMY;
        pStates[0]=pSTATE_IDLE;
        jSetSpeed*=0;
        pIndex=0;
        break;
      }
      case SDLK_1: pMaster=pMASTER_TURN_LEFT_THEN_VENTURE;  break;
      case SDLK_5: pMaster=pMASTER_VENTURE;                 break;
    	case SDLK_6: pMaster=pMASTER_DOCKING1_OLD_TURN_POINT; break;    	
    	case SDLK_9: pMaster=pMASTER_DEBUG1;                  break;
  	}
  }
  if (mySDL_Event.type==SDL_MOUSEBUTTONDOWN) {
    int pix=mySDL_Event.button.x;
    int piy=mySDL_Event.button.y;    
    if (pBAR_GAIN_INT_X<pix&&pix<pBAR_GAIN_INT_X+jBAR_W&&pBAR_GAIN_INT_Y<piy&&piy<pBAR_GAIN_INT_Y+jBAR_H)
      pGainInt=(pix-pBAR_GAIN_INT_X)*pGAININT/(float)jBAR_W;    
    if (pBAR_GAIN_POS_X<pix&&pix<pBAR_GAIN_POS_X+jBAR_W&&pBAR_GAIN_POS_Y<piy&&piy<pBAR_GAIN_POS_Y+jBAR_H)
      pGainPos=(pix-pBAR_GAIN_POS_X)*pGAINPOS/(float)jBAR_W;    
    //if (jBAR_YELLOW_X<pix&&pix<jBAR_YELLOW_X+jBAR_W&&jBAR_YELLOW_Y<piy&&piy<jBAR_YELLOW_Y+jBAR_H)
    //  jYellow=(pix-jBAR_YELLOW_X)*2.0f/(float)jBAR_W;    
  }  
}

void pAddState(int myState) {
  pStates[pIndex++]=myState;
  pStates[pIndex  ]=pSTATE_IDLE;      // just in case one forgets idle state
  if (myState==pSTATE_IDLE)
    pIndex=0;
}

void cPlanStates(void) {
  pIndex=0;
  pMasterOld=pMaster;
  switch (pMaster) {
    case pMASTER_NO_AUTONOMY: {
      pAddState(pSTATE_IDLE);
      break;
    }
    case pMASTER_TURN_LEFT_THEN_VENTURE: {    // all implemented
      cCompass.data[0]=-1;
      cCompass.data[1]= 0;
      cCompass.data[2]= 0;
      cCompass=cPos*cCompass;
      pAddState(pSTATE_ALIGN_TO_COMPASS);
      pAddState(pSTATE_SLAM_RESET);
      pAddState(pSTATE_IDLE);
      pMaster=pMASTER_VENTURE;
      break;
    }
    case pMASTER_VENTURE: {                   // all implemented
      matrix<float> dif(3,1);
      cRouting(cROUTE);
      if (cBestQuality()&&jTic[jTOC_STONES_SINCE_VENTURE]<3) {
        cPlanning();                          // provides compass
        pAddState(pSTATE_ALIGN_TO_COMPASS);
        jTog[jTOG_ABORT_PATH_ON_STONE ]=true;
        jTog[jTOG_ABORT_PATH_ON_DOCK  ]=true;
        pAddState(pSTATE_DRIVE_ALONG_PATH);        
        pAddState(pSTATE_TRACK_STONE_GRAB);                        
      } else {
        pMaster=pMASTER_DOCKING1_OLD_TURN_POINT;
      }
      pAddState(pSTATE_IDLE);
      break;
    }
    case pMASTER_DOCKING1_OLD_TURN_POINT: {         // all implemented
      jTic[jTOC_STONES_SINCE_VENTURE]=0;
      cRouting(cROUTEHOME);
      cTarget=cTurn;
      cPlanning();
      pAddState(pSTATE_ALIGN_TO_COMPASS);
      jTog[jTOG_ABORT_PATH_ON_STONE ]=false;
      jTog[jTOG_ABORT_PATH_ON_DOCK  ]=false;
      pAddState(pSTATE_DRIVE_ALONG_PATH);
      pAddState(pSTATE_IDLE);
      
      pMaster=pMASTER_DOCKING2_SAFEGUARD;
      break;
    }
    case pMASTER_DOCKING2_SAFEGUARD: {
      cRouting(cROUTE);
      cTarget=cDrop;
      cPlanning();
      pAddState(pSTATE_ALIGN_TO_COMPASS);
      
      pMaster=pMASTER_DOCKING3_KICKASS;
      break;
    }
    case pMASTER_DOCKING3_KICKASS: {
      cCompass.data[0]=cDock.data[6]-cPos.data[6];
      cCompass.data[1]=cDock.data[7]-cPos.data[7];
      cCompass.data[2]=0;
      cCompass.normalize(2);
      pAddState(pSTATE_ALIGN_TO_COMPASS);
      pAddState(pSTATE_VIVE_LA_RESISTANCE);
      pAddState(pSTATE_FUBAR);
      pAddState(pSTATE_CLEAR_PROFILE);      
      pAddState(pSTATE_IDLE);
      
      pMaster=pMASTER_VENTURE;
      break;
    }
    case pMASTER_DEBUG1: {
      //cBestQuality();
      //cPlanning();
      //pAddState(pSTATE_ALIGN_TO_COMPASS);
      //pAddState(pSTATE_DRIVE_ALONG_PATH);
      //pAddState(pSTATE_BACKUP_ON_PATH);
      pAddState(pSTATE_IDLE);
      pMaster=pMASTER_NO_AUTONOMY;
      break;
    }
  }
  jDrawString(screen,"!",7,pBAR_MASTER_Y);
  jTic[jTOC_NEXT_STRATEGY]=jTic[jTOC_LAST_SENSORS]+pTIMEOUT;
}

void pControl(void) {
  int c1,c2;
  bool next=false;
  int pState=pStates[pIndex];
  switch (pState) {
    case pSTATE_IDLE:
      cPlanStates();
      break;
    case pSTATE_ALIGN_TO_COMPASS: {
      matrix<float> loc=cInv*cCompass;
      float ang=atan2(-loc.data[0],loc.data[1]);    // angle with respect to front, left is positive
      jSetSpeed.data[1]=pGainInt*ang-.22*jVel.data[0];
      sprintf(jChar,"  c%6.3f %i",ang*jRAD2DEG,jTic[jTOC_NEXT_SUBSTANTIAL]);
      jDrawString(screen,jChar,0,pBAR_MASTER_Y+22);
      if (fabs(ang)*jRAD2DEG<7) {
        if (!jTic[jTOC_NEXT_SUBSTANTIAL])
          jTic[jTOC_NEXT_SUBSTANTIAL]=jTic[jTOC_LAST_SENSORS]+1000;
        else
          if (jTic[jTOC_NEXT_SUBSTANTIAL]<jTic[jTOC_LAST_SENSORS]) {
            jTic[jTOC_NEXT_SUBSTANTIAL]=0;
            jSetSpeed.data[1]=0;     
            next=true;
          }
      }
      break;
    }
    case pSTATE_DRIVE_ALONG_PATH: {
      int pin;
      matrix<float> loc(3,1);
      while (cPlan.next[0]) {
        pin=cPlan.next[0]->data;
        loc.data[0]=((pin/640)-jMAP_X)*jMAP2GND;
        loc.data[1]=((pin%640)-jMAP_X)*jMAP2GND;
        loc.data[2]=1;
        loc=cInv*loc;
        loc.data[2]=0;
        if (loc.norm(2)<0.30f) {
          cPlan.next[2]=cPlan.next[0];  // TEST WITHGOUT THIS LINE!!!
          cPlan.remove();
        } else
          break;
      }
      cPlan.next[2]=&cPlan; // should reset tokenizer // !cPathHazardous()
      if (cPlan.next[0]) {  // end of plan is reached
        float ang=atan2(-loc.data[0],loc.data[1]);  // angle with respect to front, left is positive
        //jSetSpeed.data[0]=0.35/(1.0f+fabs(ang/5.0f));   // this is optimized for use with odometry
        jSetSpeed.data[0]=jSPEED_HED/(1.0f+3*fabs(ang));
        jSetSpeed.data[1]=pGainInt*ang-pGainPos*jVel.data[0];
      } else {
        jTog[jTOG_ABORT_PATH_ON_STONE]=false;       // signal: no stone found
        jTog[jTOG_ABORT_PATH_ON_DOCK ]=false;       // signal: no stone found
        jSetSpeed*=0;
        next=true;
      }
      if (jTog[jTOG_ABORT_PATH_ON_DOCK ]&&jTog[jTOG_DOCK_IN_SIGHT]) {
        jTog[jTOG_ABORT_PATH_ON_STONE]=false;       // signal: no stone found
        jTog[jTOG_ABORT_PATH_ON_DOCK ]=false;       // signal: no stone found
        jSetSpeed*=0;
        next=true;
      }
      if (jTog[jTOG_ABORT_PATH_ON_STONE]&&jReachable.size[1]) {
        matrix<float> nrm(2,1);
        matrix<float> loc=cInv*cDrop;               // dock in local coordinates
        for (c1=0;c1<jReachable.size[1];c1++) {
          nrm.data[0]=jReachable.data[c1*3+0]-loc.data[6];
          nrm.data[1]=jReachable.data[c1*3+1]-loc.data[7];
          if (0.5f<nrm.norm(2)) {                   // stone is far away from dock
            matrix<float> tmp(3,3,"eye");
            tmp.data[6]=jReachable.data[c1*3+0];
            tmp.data[7]=jReachable.data[c1*3+1];
            cTarget=cPos*tmp;
            jSetSpeed*=0;
            next=true;
            break;
          }
        }
      }
      break;
    }
    case pSTATE_TRACK_STONE_GRAB: {
      if (jTog[jTOG_ABORT_PATH_ON_STONE]) {
        pTargetStone();
        matrix<float> loc(3,1);       // contains target in local coordinates
        loc.data[0]=cTarget.data[6];  // target to the left
        loc.data[1]=cTarget.data[7];  // target to the front
        loc.data[2]=1;
        loc=cInv*loc;                 // contains target in local coordinates
        loc.data[2]=0;
        float ang=atan2(-loc.data[0],loc.data[1]);  // angle with respect to front, left is positive
        float dst=loc.norm(2);
        if (0.45f<dst&&fabs(ang)<60*jDEG2RAD) {
          jSetSpeed.data[0]=MAX(0.15f-fabs(ang)*0.3f,0);
          jSetSpeed.data[1]=pGainInt*ang-0.35f*jVel.data[0];
        } else {
          jTic[jTOC_STONES_SINCE_VENTURE]++;
          jSetSpeed*=0;
          next=true;
        }
      } else {
        jSetSpeed*=0;
        next=true;
      }
      break;
    }
    case pSTATE_VIVE_LA_RESISTANCE: {
      matrix<float> loc(3,1);       // contains target in local coordinates
      loc.data[0]=cDock.data[6];    // target to the left
      loc.data[1]=cDock.data[7];    // target to the front
      loc.data[2]=1;
      loc=cInv*loc;                 // contains target in local coordinates
      loc.data[2]=0;
      float ang=atan2(-loc.data[0],loc.data[1]);  // angle with respect to front, left is positive
      float dst=loc.norm(2);
      jSetSpeed.data[0]=MAX(0.15f-fabs(ang)*0.3f,0);
      jSetSpeed.data[1]=pGainInt*ang-0.35f*jVel.data[0];
      sprintf(jChar,"  |%6.3f a%6.3f v%6.3f resist",dst,ang*jRAD2DEG,jVel.norm(2));
      jDrawString(screen,jChar,0,pBAR_MASTER_Y+22);
      if (!jTic[jTOC_NEXT_SUBSTANTIAL])
        jTic[jTOC_NEXT_SUBSTANTIAL]=jTic[jTOC_LAST_SENSORS]+10000;
      else
        if (jTic[jTOC_NEXT_SUBSTANTIAL]<jTic[jTOC_LAST_SENSORS] || dst<0.60f ) {
          jTic[jTOC_NEXT_SUBSTANTIAL]=0;
          jSetSpeed*=0;
          next=true;
        }
      break;
    }
    case pSTATE_FUBAR: {
      jSetSpeed.data[0]=-0.15f;
      jSetSpeed.data[1]= 0    ;
      if (!jTic[jTOC_NEXT_SUBSTANTIAL])
        jTic[jTOC_NEXT_SUBSTANTIAL]=jTic[jTOC_LAST_SENSORS]+3500;
      else
        if (jTic[jTOC_NEXT_SUBSTANTIAL]<jTic[jTOC_LAST_SENSORS]) {
          jTic[jTOC_NEXT_SUBSTANTIAL]=0;
          jSetSpeed*=0;
          next=true;
        }
      break;
    }    
    case pSTATE_SLAM_RESET: {
      cReset();
      jSetSpeed*=0;
      next=true;
      break;
    }
    case pSTATE_CLEAR_PROFILE: {
      jTic[jTOC_STONES_SINCE_VENTURE]=0;
      SDL_FillRect(cProfileDraw,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
      SDL_FillRect(cProfileGone,NULL,SDL_MapRGBA(&jSDL_PixelFormat,000,000,000,000));
      jSetSpeed*=0;
      next=true;
      break;
    }
  }
  
  int cnt=0;
  while (pStates[cnt]) {
    if (cnt==pIndex)
      sprintf(jChar,">%i",pStates[cnt]);
    else
      sprintf(jChar," %i",pStates[cnt]);
    jDrawString(screen,jChar,cnt++*21+7,pBAR_MASTER_Y+11);
  }
  
  if (jTic[jTOC_NEXT_STRATEGY]<jTic[jTOC_LAST_SENSORS]) {
    pIndex=0;
    pStates[0]=pSTATE_IDLE;
    jTic[jTOC_NEXT_SUBSTANTIAL]=0;
    jSetSpeed*=0;
    next=false;
  }
  if (next)
    pIndex++;
}

void pHUD(void) {
  SDL_Rect mySDL_src;
  SDL_Rect mySDL_dst;
  mySDL_src.x=0; mySDL_src.y=0; mySDL_src.h=jBAR_H;
  jDrawString(screen,"heading",0,pBAR_GAIN_INT_Y);
  // draws the bar: pGainInt
  mySDL_src.w=(int)(jBAR_W*pGainInt/pGAININT);
  mySDL_dst.x=pBAR_GAIN_INT_X;
  mySDL_dst.y=pBAR_GAIN_INT_Y;
  SDL_BlitSurface(jBarT,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%4.2f",pGainInt);
	jDrawString(screen,jChar,pBAR_GAIN_INT_X+jBAR_W/2-7,pBAR_GAIN_INT_Y);
  // draws the bar: pGainInt
  mySDL_src.w=(int)(jBAR_W*pGainPos/pGAINPOS);
  mySDL_dst.x=pBAR_GAIN_POS_X;
  mySDL_dst.y=pBAR_GAIN_POS_Y;
  SDL_BlitSurface(jBarV,&mySDL_src,screen,&mySDL_dst);
	sprintf(jChar,"%4.2f",pGainPos);
	jDrawString(screen,jChar,pBAR_GAIN_POS_X+jBAR_W/2-7,pBAR_GAIN_POS_Y);
  sprintf(jChar,"? m%i -> m%i s%i",pMasterOld,pMaster,pStates[pIndex]);
	jDrawString(screen,jChar,0,pBAR_MASTER_Y);
  // draws the bar: pGainInt
  mySDL_src.w=MAX(0,jTic[jTOC_NEXT_STRATEGY]-jTic[jTOC_LAST_SENSORS])*jBAR_W/pTIMEOUT;
  mySDL_dst.x=pBAR_MASTER_X;
  mySDL_dst.y=pBAR_MASTER_Y;
  SDL_BlitSurface(jBarB,&mySDL_src,screen,&mySDL_dst);
}
