#include "j2b2GIMICtrl.hpp"
#include "owndebug.h"

#include <SDL/SDL.h>
#include <SDL/SDL_gfxPrimitives.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_thread.h>

using namespace J2B2GIMICtrl;
static volatile bool run=true;

static std::string own_group_id="real";

SDL_Surface *screen=NULL;
SDL_Surface *camera[2];

ASRobo::Drivers::Devices::Ranging::CDistanceArray laser;
ASRobo::Drivers::Devices::Ranging::CDistanceArray bumper;
gim::time timestamp;
CJ2B2GIMICtrl ctrl;

#define jSCREEN_W         1020
#define jSCREEN_H         880

#include <math.h>
#include "inc/ntree.hpp"
#include "inc/array.hpp"
#include "inc/matrix.hpp"
#include "inc/jdraw.h"
#include "inc/jcode.cpp"
#include "inc/ccode.cpp"
#include "inc/pcode.cpp"
#include "inc/jdraw.c"

int main(int argc, char *argv[]) {
  int c1;
  bool result;  
  debugInit();
  debugSetLogFilename("j2b2gimictrl_example.log");
  debugSetPrintLevel(3);
  debugSetLogLevel(3);
  result=ctrl.Connect("turbo.dyndns.tv",51002,"j2b2ctrl_test_"+own_group_id);
  if (!result) {
    dPrint(1,"Failed to Connect J2B2GIMICtrl. Fatal error - bailing out.");
    exit(1);
  }
  result=ctrl.InitializeASRobo("ASRobo_"+own_group_id);
  if (!result) {
    dPrint(1,"Failed to Initialize ASRobo!!! Operating without it then.");
    ownSleep_ms(2000);
    exit(1);
  }  
  result=ctrl.InitializePTU("PTU_"+own_group_id);
  if (!result) {
    dPrint(1,"Warning: Failed to Initialize PTU!!! Operating without it then.");
    ownSleep_ms(2000);
    exit(1);
  }
  result=ctrl.InitializeImageClient("ImageServer_"+own_group_id);
  if (!result) {
    dPrint(1,"Warning: Failed to Initialize ImageClient!!! Operating without it then.");
    ownSleep_ms(2000);
    exit(1);
  }
  if (SDL_Init(SDL_INIT_VIDEO|SDL_INIT_JOYSTICK)==-1) {
    dPrint(1,"Failed to initialize SDL (Video & Joystick): %s\n",SDL_GetError());
    exit(1);    
  }
  atexit(SDL_Quit);
  screen=SDL_SetVideoMode(jSCREEN_W,jSCREEN_H,24,SDL_SWSURFACE|SDL_DOUBLEBUF);  
  SDL_WM_SetIcon(SDL_LoadBMP("res/jicon.bmp"),NULL);
  SDL_WM_SetCaption("Group1 :: 07dec2007",NULL);
  dPrint(1,"ASRobo - Drive - SetAcceleration to 0.2m/s");
  ctrl.ASRobo_SetAcceleration(jAccel);  // Set acceleration to 0.2m/s
  int laserseq=-1;  // IMPORTANT, DON'T ALTER
  int bumperseq=-1; // IMPORTANT, DON'T ALTER
  
  ctrl.ASRobo_SetAllBehaviours(false);
  ctrl.ASRobo_IO_SetPin(ASROBO_IOPIN_LIGHT,true);
  ctrl.ASRobo_IO_SetPin(ASROBO_IOPIN_LASER,true);
  
  jInit();
  cInit();
  pInit();
  
  dPrint(1,"Entering SDL Event Processing loop... Press 'ESC' to exit!");
  
  TBitmap myTBitmap;
  SDL_RWops *mySDL_RWops;
  SDL_Surface *mySDL_Surface;
  SDL_Event event;
  
  while (run) {
    // BEGIN: query sensors into global variables
    camera[0]=NULL;
    camera[1]=NULL;
    for (c1=1;c1<2;c1++) {
      camera[c1]=NULL;
      if (ctrl.IC_GetImage(c1,&myTBitmap)) {
        if (mySDL_RWops=SDL_RWFromMem(myTBitmap.data,myTBitmap.data_size)) {
          if (mySDL_Surface=IMG_LoadJPG_RW(mySDL_RWops)) {
            camera[c1]=SDL_ConvertSurface(mySDL_Surface,&jSDL_PixelFormat,0);
            SDL_FreeSurface(mySDL_Surface);
          }
          SDL_FreeRW(mySDL_RWops);
        }
        myTBitmap.free_data();
      }
    }
    ctrl.ASRobo_GetLaserScan(laser,&laserseq);
    ctrl.ASRobo_GetBumperScan(bumper,&bumperseq);
    
    if (laser.distanceArray.size()==jLaserSize) {
    // END: query sensors into global variables
    
    // BEGIN: Instantaneous perception
    jSensors();           // state matrix, laser coordinates, camera precomputations        
    jStones();            // stone detection    
    jReachables();
    jDocks();             //  dock detection
    // END: Instantaneous perception
    
    // BEGIN: Permanent conception    
    cProfile();           // maps stones and takes picture of ground    
    cLocalize();          // determines position
    cJustify();           // performs numerical correction
    cMapping();           // updates map
    // END: Permanent conception
    
    // BEGIN: Instantaneous perception
    jHazards();           // measures if hazard is detected, uses most recent speed
    // END: Instantaneous perception
    
    while (SDL_PollEvent(&event)) {
      
      jProcess(event);    // camera pan and tilt
      cProcess(event);    // 
      pProcess(event);    // 
      
      switch (event.type) {
        case SDL_QUIT: {
	        run=false;
	        break;
        }
        case SDL_KEYDOWN: {
        	switch (event.key.keysym.sym) {
    	      case SDLK_ESCAPE:
    	        run=false;
    	        break;
          }
	        break;
        }
      }
    }
    
    pControl();           // strategy
    jControl();           // basic behavior and control of actuators
    
    jCameraPlot();        // lines of straight ahead
    jFrontPlot();         // contributes to front view: circle of robot, radii of distance
    cMapPlot();           // contributes to map
    jHUD();
    cHUD();
    pHUD();
    jDraw();              // draws cameras, front view, and map

    SDL_Flip(screen);
    SDL_FillRect(screen,NULL,SDL_MapRGBA(screen->format,255,255,255,0));
    
    if (camera[0])
      SDL_FreeSurface(camera[0]);
    if (camera[1])
      SDL_FreeSurface(camera[1]);    
    ownSleep_ms(5);
    }
  }
  dPrint(1,"Exit.");
  ownSleep_ms(50); // Exit program.
  return 0;
}
